import importlib

from importlib_metadata import entry_points

registered_plugins = []


def register_plugins():
    """Register a plugin."""
    # Find all plugins that defined an entry point
    discovered_plugins = entry_points(group="octarine.plugins")

    # Go over each of the plugins
    for plugin in discovered_plugins:
        # Import the module
        try:
            module = importlib.import_module(plugin.module)
        except BaseException as e:
            print(f"Error importing plugin {plugin.name}: {e}")
            continue

        # Get the function to register the plugin
        register_func = getattr(module, plugin.value.split(":")[-1], None)

        # If the function is not found, print an error
        if register_func is None:
            print(
                f"Registration function {plugin.value.split(':')[-1]} not found for plugin {plugin.name}."
            )
            continue

        # Otherwise, register the plugin
        try:
            register_func()
        except BaseException as e:
            print(f"Error registering plugin {plugin.name}: {e}")

        # Add the plugin to the list of registered plugins
        registered_plugins.append(plugin.name)


register_plugins()



import pygfx as gfx
import trimesh as tm

from .utils import is_hashable, is_points, is_lines, is_volume, is_mesh_like, is_pygfx_visual, is_pygfx_geometry
from .visuals import points2gfx, lines2gfx, mesh2gfx, trimesh2gfx, volume2gfx, scene2gfx, geometry2gfx, visual_passthrough

CONVERTERS = {
    is_pygfx_visual: visual_passthrough,  # pass-through
    is_pygfx_geometry: geometry2gfx,  # add default material and return
    tm.Trimesh: trimesh2gfx,
    tm.Scene: scene2gfx,
    is_mesh_like: mesh2gfx,
    is_points: points2gfx,
    is_lines: lines2gfx,
    is_volume: volume2gfx
}


def get_converter(t, raise_missing=True):
    """Get the converter for a given data type."""
    # Go through converters in order
    for k, v in CONVERTERS.items():
        # First check if we have a direct match
        if is_hashable(t) and t == k:
            return v

        # If not, check if k is a type
        if type(t) == k:
            return v

        # Check if t is a subclass of k
        if isinstance(k, type) and isinstance(t, k):
            return v

        # Check if k is a callable
        if callable(k):
            try:
                if k(t) is True:
                    return v
            except Exception:
                pass

    if not raise_missing:
        return None

    raise NotImplementedError(f"No converter found for {t} ({type(t)}).")


def register_converter(t, converter, insert='first'):
    """Register a converter for a given data type.

    Parameters
    ----------
    t :         type | hashable | callable
                Data type the converter is meant to convert. If a function
                it is expected to take a single argument `x` and return
                True if `x` can be converted using `converter` and False
                if not.
    converter : callable
                Function that converts `x` to pygfx visuals. Must accept
                at least a single argument and return either a single
                visual or a list thereof.
    insert :    "first" | "last"
                Whether to insert the converter at the beginning or end
                of the list of converters. This is important because when
                looking for a converter for a given type we will use the
                first one that matches.

    """
    global CONVERTERS
    assert insert in ('first', 'last')

    if not callable(converter):
        raise ValueError("Converter must be callable.")

    if not callable(t) and not is_hashable(t):
        raise ValueError("Type must be hashable or callable.")


    if insert == 'first':
        items = list(CONVERTERS.items())
        items.insert(0, (t, converter))
        CONVERTERS = dict(items)
    else:
        CONVERTERS[t] = converter


import logging
import os

HEADLESS = False
PRIMARY_VIEWER = None

logger = logging.getLogger('octarine')


def default_logging():
    """Add a formatted stream handler to the ``octarine`` logger.

    Called by default when octarine is imported for the first time.
    To prevent this behaviour, set an environment variable:
    ``OCTARINE_SKIP_LOG_SETUP=True``.
    """
    logger.setLevel(logging.INFO)
    if len(logger.handlers) == 0:
        sh = logging.StreamHandler()
        sh.setLevel(logging.DEBUG)
        # Create formatter and add it to the handlers
        formatter = logging.Formatter(
            '%(levelname)-5s : %(message)s (%(name)s)')
        sh.setFormatter(formatter)
        logger.addHandler(sh)


def remove_log_handlers():
    """Remove all handlers from the ``octarine`` logger.

    It may be preferable to skip octarine's default log handler
    being added in the first place.
    Do this by setting an environment variable before the first import:
    ``OCTARINE_SKIP_LOG_SETUP=True``.
    """
    logger.handlers.clear()


skip_log_setup = os.environ.get('OCTARINE_SKIP_LOG_SETUP', '').lower() == 'true'
if not skip_log_setup:
    default_logging()


def get_logger(name: str):
    if skip_log_setup:
        return logging.getLogger(name)
    return logger


import pygfx as gfx

from ipywidgets import VBox
from IPython.display import display

from ipywidgets.widgets import (
    SelectMultiple,
    FloatSlider,
    ColorPicker,
    Checkbox,
    HBox,
    Tab,
    Dropdown,
    Layout,
    Button,
)

try:
    from sidecar import Sidecar
except ImportError:
    Sidecar = None


class JupyterOutput(HBox):
    """Wrap Viewer (and potentially a toolbar) in Vbox for display."""

    def __init__(self, viewer, toolbar: bool, use_sidecar: bool, sidecar_kwargs: dict):
        """

        Parameters
        ----------
        viewer  :       Viewer
                        The Viewer instance to display.
        toolbar :       bool
                        Whether to display a toolbar. If False, the toolbar
                        will still be created but will hidden.
        use_sidecar :   bool
                        Whether to use Sidecar for display. Will throw an
                        error if Sidecar is not installed.
        sidecar_kwargs: dict
                        optional kwargs passed to Sidecar

        """
        self.viewer = viewer
        self.sidecar = None

        self.use_sidecar = use_sidecar

        if self.use_sidecar and Sidecar is None:
            raise ImportError(
                "Sidecar extention is not installed: `pip install sidecar`\n"
                "Note that you will have to restart Jupyter and deep-reload "
                "the page after installation."
            )

        self.toolbar = JupyterToolbar(viewer)
        self.output = (
            viewer.canvas,
            self.toolbar,
        )

        if not toolbar:  # just stack canvas in VBox
            self.toolbar.hide()

        if use_sidecar:  # instantiate sidecar if requested
            self.sidecar = Sidecar(**sidecar_kwargs)

        # stack all objects in the HBox
        super().__init__(self.output)

    def _repr_mimebundle_(self, *args, **kwargs):
        """
        This is what jupyter hook into when this output context instance is returned at the end of a cell.
        """
        if self.use_sidecar:
            with self.sidecar:
                return display(VBox(self.output))
        else:
            # just display VBox contents in cell output
            return super()._repr_mimebundle_(*args, **kwargs)

    def close(self, close_viewer=True):
        """Closes the output context, cleanup all the stuff"""
        self._is_closed = True  #  we need this to avoid recursion in close

        if close_viewer:
            self.viewer.close()

        if self.toolbar is not None:
            self.toolbar.close()

        if self.sidecar is not None:
            self.sidecar.close()

        super().close()  # ipywidget VBox cleanup


class JupyterToolbar(VBox):
    """Basic toolbar using ipywidgets"""

    def __init__(self, viewer, width="150px"):
        self.viewer = viewer

        self.legend = SelectMultiple(
            options=[*"abcdefghijklmonp"],
            value=[],
            # rows=10,
            description="",
            disabled=False,
            layout=Layout(
                width="auto", height="200px"
            ),  # make sure this doesn't take too much space
        )
        self.update_legend()  # update legend entries
        self.color_picker = ColorPicker(
            concise=False,
            description="",
            value="white",
            disabled=False,
            layout=Layout(width="auto"),
            tooltip="Set color for selected objects",
        )
        self.actions_dropdown = Dropdown(
            options=["Action", "Remove selected", "Hide selected", "Unhide all"],
            value="Action",
            description="",
            disabled=False,
            layout=Layout(width="auto"),
            tooltip="Perform action on selected objects",
        )
        self.center_camera_button = Button(
            description="Center Camera",
            value=False,
            disabled=False,
            icon="align-center",
            layout=Layout(width="auto"),
            tooltip="Auto-center camera",
        )
        self.screenshot_button = Button(
            description="Screenshot",
            value=False,
            disabled=False,
            icon="camera",
            layout=Layout(width="auto"),
            tooltip="Take a screenshot",
        )
        self.flat_shading_toggle = Checkbox(
            value=False,
            description="Flat shading",
            disabled=False,
            indent=False,
            layout=Layout(width="auto"),
        )
        self.wireframe_toggle = Checkbox(
            value=False,
            description="Wireframe",
            disabled=False,
            indent=False,
            layout=Layout(width="auto"),
        )
        self.ambient_light_toggle = Checkbox(
            value=True,
            description="Ambient Light",
            disabled=False,
            indent=False,
            layout=Layout(width="auto"),
        )
        self.ambient_light_slider = FloatSlider(
            value=0.5,
            min=0,
            max=5,
            step=0.1,
            description="",
            disabled=False,
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format=".1f",
            layout=Layout(width="auto"),
        )
        self.bounds_toggle = Checkbox(
            value=False,
            description="Show Bounds",
            disabled=False,
            indent=False,
            layout=Layout(width="auto"),
        )

        # self._panzoom_controller_button.observe(self.panzoom_handler, "value")
        self.center_camera_button.on_click(self.center_camera)
        self.screenshot_button.on_click(self.screenshot)
        self.legend.observe(self.object_selected, names="value")
        self.color_picker.observe(self.change_color, names="value")
        self.actions_dropdown.observe(self.action_selected, names="value")
        self.flat_shading_toggle.observe(self.set_shading, names="value")
        self.wireframe_toggle.observe(self.set_wireframe, names="value")
        self.ambient_light_toggle.observe(self.toggle_ambient_light, names="value")
        self.ambient_light_slider.observe(self.set_ambient_light, names="value")
        self.bounds_toggle.observe(self.toggle_bounds, names="value")

        # Split into object and scene widgets
        object_widgets = [
            self.legend,
            self.actions_dropdown,
            self.color_picker,
            self.center_camera_button,
            self.screenshot_button,
        ]
        scene_widgets = [
            self.flat_shading_toggle,
            self.wireframe_toggle,
            self.ambient_light_toggle,
            self.ambient_light_slider,
            self.bounds_toggle,
        ]

        # Combine object and scene widgets in tabs
        self.tabs = Tab()
        self.tabs.children = [
            VBox(object_widgets, layout=Layout(width="auto")),
            VBox(scene_widgets, layout=Layout(width="auto")),
        ]
        self.tabs.titles = ["Objects", "Scene"]

        super().__init__([self.tabs], layout=Layout(width=width))

    def hide(self):
        """Hide the toolbar."""
        self.layout.display = "none"

    def show(self):
        """Show the toolbar."""
        self.layout.display = "block"

    def toggle(self):
        """Toggle visibility of the toolbar."""
        if self.layout.display == "none":
            self.show()
        else:
            self.hide()

    def center_camera(self, obj):
        """Center camera on scene."""
        self.viewer.center_camera()

    def screenshot(self, obj):
        pass

    def update_legend(self):
        """Update entries in the legend widget."""
        self.legend.options = list(self.viewer.objects.keys())

    def object_selected(self, change):
        """This is triggered when an object is selected in the legend widget."""
        # Change is a tuple of (item1, item2, ..) from the legend widget
        if len(change["new"]) == 1:
            # Get this object
            ob = self.viewer[change["new"][0]][0]
            # Get the color and assign to the color picker widget
            self.color_picker.value = ob.material.color.hex

    def action_selected(self, change):
        """Perform an action on selected objects."""
        # Reset the button value to "Action"
        self.actions_dropdown.value = "Action"

        sel = self.legend.value
        if len(sel):
            if change["new"] == "Remove selected":
                self.viewer.remove_objects(sel)
            elif change["new"] == "Hide selected":
                self.viewer.hide_objects(sel)
            elif change["new"] == "Unhide all":
                self.viewer.unhide_objects()

    def change_color(self, change):
        """This is triggered when the color picker is changed."""
        self.viewer.set_colors({ob: change["new"] for ob in self.legend.value})

    def set_shading(self, change):
        """Set flat shading for all meshes."""
        for ob in self.viewer.scene.children:
            if isinstance(ob, gfx.Mesh):
                ob.material.flat_shading = change["new"]

    def set_wireframe(self, change):
        """Set wireframe mode for all meshes."""
        for ob in self.viewer.scene.children:
            if isinstance(ob, gfx.Mesh):
                ob.material.wireframe = change["new"]

    def set_ambient_light(self, change):
        """Set intensity of ambient lights."""
        for ob in self.viewer.scene.children:
            if isinstance(ob, gfx.AmbientLight):
                ob.intensity = change["new"]

    def toggle_ambient_light(self, change):
        """Toggle visibility of ambient lights."""
        for ob in self.viewer.scene.children:
            if isinstance(ob, gfx.AmbientLight):
                ob.visible = change["new"]

    def toggle_bounds(self, change):
        """Toggle visibility of scene bounds."""
        self.viewer.show_bounds = change["new"]



import os
import png
import sys
import time
import cmap
import uuid
import random
import inspect
import warnings

import numpy as np
import pygfx as gfx
import trimesh as tm

from pathlib import Path
from collections import OrderedDict
from functools import wraps, lru_cache, partial
from pygfx.renderers.wgpu.engine.edl import EDLPass

from rendercanvas.offscreen import OffscreenRenderCanvas

from .visuals import mesh2gfx, volume2gfx, points2gfx, lines2gfx, text2gfx
from .conversion import get_converter
from . import utils, config


__all__ = ["Viewer", "viewers"]

logger = config.get_logger(__name__)

# This keeps track of open viewers
viewers = []

AUTOSTART_EVENT_LOOP = True

EFFECT_CLASSES = {
    "edl": EDLPass,
}

# TODO
# - add styles for viewer (lights, background, etc.) - e.g. .set_style(dark)
#   - e.g. material.metalness = 2 looks good for background meshes
#   - metalness = 1 with roughness = 0 makes for funky looking neurons
#   - m.material.side = "FRONT" makes volumes look better
# - make Viewer reactive (see reactive_rendering.py) to save
#   resources when not actively using the viewer - might help in Jupyter?
# [/] add specialised methods for adding neurons, volumes, etc. to the viewer
# - move lights to just outside the scene's bounding box (maybe use decorator?)
#   whenever we add/remove objects


def update_viewer(legend=True, bounds=True):
    def outer(func):
        """Decorator to update legend and other properties."""

        @wraps(func)
        def inner(*args, **kwargs):
            # Run function first
            func(*args, **kwargs)
            update_helper(viewer=args[0], legend=legend, bounds=bounds)

        return inner

    return outer


def update_helper(viewer, legend=True, bounds=True):
    """Helper function to update legend and other properties."""
    # Always clear the cached objects dictionary
    viewer._objects.cache_clear()

    if legend:
        if getattr(viewer, "controls", None):
            viewer.controls.update_legend()
        if getattr(viewer, "widget", None):
            if viewer.widget.toolbar:
                viewer.widget.toolbar.update_legend()
    if bounds:
        if getattr(viewer, "show_bounds", False):
            viewer.update_bounds()

    # Any time we update the viewer, we should set it to stale
    viewer._render_stale = True
    viewer.canvas.request_draw()


class Viewer:
    """PyGFX 3D viewer.

    Parameters
    ----------
    offscreen : bool
                If True, will use an offscreen Canvas. Useful if you only
                want a screenshot.
    title :     str
                Title of the viewer window.
    max_fps :   int
                Maximum frames per second to render.
    size :      tuple, optional
                Size of the viewer window.
    camera :    "ortho" | "perspective"
                Type of camera to use. Defaults to "ortho". Note you can always
                change the camera type by adjust the `Viewer.camera.fov` attribute
                (0 = ortho, >0 = perspective).
    control :   "trackball" | "panzoom" | "fly" | "orbit"
                Controller type to use. Defaults to "trackball".
    show :      "auto" (default) | bool
                Whether to immediately show the viewer. When set to "auto" (default),
                will immmediately show the viewer if:
                 - we are in a Jupyter environment
                 - we are in an iPython session and we can hook into an iPython event loop
                If neither of the above applies or `show=False`, you will have to manually run
                `Viewer.show()`. This gives you the chance to add objects to the viewer
                before it is shown and the blocking event loop is started.
                The `show` parameter is ignored if `offscreen` is True.
    **kwargs
                Keyword arguments are passed through to ``WgpuCanvas``.

    """

    # Palette used for assigning colors to objects
    palette = "seaborn:tab10"
    highlight_color = "yellow"

    def __init__(
        self,
        offscreen=False,
        title="Octarine Viewer",
        max_fps=30,
        camera="ortho",
        control="trackball",
        size=None,
        show=True,
        **kwargs,
    ):
        # We need to import WgpuCanvas before we (potentially) start the event loop
        # If we don't we get a segfault.
        if not offscreen:
            from rendercanvas.auto import RenderCanvas

        # Check if we're running in an IPython environment
        if utils._type_of_script() == "ipython" and not offscreen:
            ip = get_ipython()  # noqa: F821
            if not ip.active_eventloop:
                if AUTOSTART_EVENT_LOOP:
                    try:
                        ip.enable_gui("qt6")
                        logger.debug(
                            "Looks like you're running in an IPython environment but haven't "
                            "started a GUI event loop. We've started one for you using the "
                            "Qt6 backend."
                        )
                    except (ModuleNotFoundError, ImportError):
                        raise ValueError(
                            "Looks like you're running in an IPython environment but haven't "
                            "started a GUI event loop. We tried to start one for you using the "
                            "Qt6 backend (via %gui qt6) but that failed. You may have to start "
                            "the event loop manually. See "
                            "https://ipython.readthedocs.io/en/stable/config/eventloops.html"
                            "for details."
                        )
                else:
                    raise ValueError(
                        'IPython event loop not running. Please use e.g. "%gui qt6" to hook into the event loop.'
                    )

        self._title = title

        # Update some defaults as necessary
        defaults = {"title": title, "max_fps": max_fps}
        if size is not None:
            defaults["size"] = size
        defaults.update(kwargs)

        # If we're running in headless mode (primarily for tests on CI) we will
        # simply not initialize the gfx objects. Not ideal but it turns
        # out to be very annoying to correctly setup on Github Actions.
        if getattr(config, "HEADLESS", False):
            return

        if not offscreen:
            self.canvas = RenderCanvas(**defaults)
        else:
            self.canvas = OffscreenRenderCanvas(**defaults)

        # There is a bug in pygfx 0.1.18 that causes the renderer to crash
        # when using a Jupyter canvas without explicitly setting the pixel_ratio.
        # This is already fixed in main but for now:
        if self._is_jupyter:
            self.renderer = gfx.renderers.WgpuRenderer(
                self.canvas, show_fps=False, pixel_ratio=2
            )
        else:
            self.renderer = gfx.renderers.WgpuRenderer(self.canvas, show_fps=False)

        # Set up a default scene
        self.scene = gfx.Scene()

        # A minor ambient light
        self.scene.add(gfx.AmbientLight(intensity=0.5))

        # A strong point light form front/top/left
        self.scene.add(gfx.PointLight(intensity=4))
        self.scene.children[-1].shadow.bias = 0.0000005  # this helps with shadow acne
        self.scene.children[-1].local.x = -1000000  # move to the left
        self.scene.children[-1].local.y = -1000000  # move up
        self.scene.children[-1].local.z = -1000000  # move light forward

        # A weaker point light from the back
        self.scene.add(gfx.PointLight(intensity=1))
        self.scene.children[-1].shadow.bias = 0.0000005  # this helps with shadow acne
        self.scene.children[-1].local.x = 1000000  # move to the left
        self.scene.children[-1].local.y = 1000000  # move up
        self.scene.children[-1].local.z = 1000000  # move light forward

        # Set up a default background
        self._background = gfx.Background(None, gfx.BackgroundMaterial((0, 0, 0)))
        self.scene.add(self._background)

        # Add camera
        if camera == "ortho":
            self.camera = gfx.OrthographicCamera()
        elif camera == "perspective":
            self.camera = gfx.PerspectiveCamera()
        else:
            raise ValueError(f"Unknown camera type: {camera}")

        # Add controller
        controller = {
            "trackball": gfx.TrackballController,
            "panzoom": gfx.PanZoomController,
            "fly": gfx.FlyController,
            "orbit": gfx.OrbitController,
        }.get(control, None)
        if controller is None:
            raise ValueError(f"Unknown controller type: {control}")

        self.controller = controller(self.camera, register_events=self.renderer)

        # Setup overlay
        self.overlay_camera = gfx.NDCCamera()
        self.overlay_scene = gfx.Scene()

        # Setup transform gizmo
        self.transform_gizmo = None

        # Stats
        self.stats = gfx.Stats(self.renderer)
        self._show_fps = False

        # Setup key events
        self._key_events = {}
        self._key_events["1"] = lambda: self.set_view("XY")  # frontal view
        self._key_events["2"] = lambda: self.set_view("XZ")  # lateral view
        self._key_events["3"] = lambda: self.set_view("YZ")  # top view
        self._key_events[("1", ("Shift",))] = lambda: self.set_view("-XY")  # back view
        self._key_events[("2", ("Shift",))] = lambda: self.set_view(
            "-XZ"
        )  # other lateral view
        self._key_events[("3", ("Shift",))] = lambda: self.set_view(
            "-YZ"
        )  # bottom view
        self._key_events["f"] = lambda: self._toggle_fps()
        self._key_events["c"] = lambda: self._toggle_controls()

        def _keydown(event):
            """Handle key presses."""
            if not event.modifiers:
                if event.key in self._key_events:
                    self._key_events[event.key]()
            else:
                tup = (event.key, tuple(event.modifiers))
                if tup in self._key_events:
                    self._key_events[tup]()

        # Register events
        self.renderer.add_event_handler(_keydown, "key_down")

        # Finally, setting some variables
        self._show_bounds = False
        self._shadows = False
        self._animations = {}
        self._animations_flagged_for_removal = []
        self._animations_frame_counter = 0
        self._on_double_click = None
        self._on_hover = None
        self._objects_pickable = False
        self._selected = []
        self._render_trigger = "continuous"

        viewers.append(self)

        # This starts the animation loop
        if show and not self._is_jupyter:
            self.show(start_loop=show == "start_loop")

    def _animate(self):
        """Run the rendering loop."""
        rm = self.render_trigger

        # First run the user animations
        self._animations_frame_counter += 1
        if self._animations_frame_counter == sys.maxsize:  # reset to avoid overflow
            self._animations_frame_counter = 0
        # N.B. we're iterating over the list because the user might add / remove
        # animations during the loop
        for i, (func, (on_error, run_every, req_render)) in enumerate(
            list(self._animations.items())
        ):
            # Skip if we're not supposed to run this frame
            if run_every and (self._animations_frame_counter % run_every) != 0:
                continue
            try:
                func()
                if req_render:
                    self._render_stale = True
            except BaseException as e:
                if on_error == "raise":
                    raise e
                elif on_error == "log":
                    logger.error(f"Error in animation function '{func}': {e}")
                elif on_error == "remove":
                    logger.error(
                        f"Removing animation function '{func}' because of error: {e}"
                    )
                    # Flag animation for removal
                    self._animations_flagged_for_removal.append(func)

        # Check if any animations need to be removed
        for f in self._animations_flagged_for_removal:
            try:
                _ = self._animations.pop(f)
            except KeyError:
                pass  # already removed (e.g. by index or by function)
        self._animations_flagged_for_removal = []

        # Now check if we need to render the scene
        if rm == "active_window":
            # Note to self: we need to explore how to do this with different backends / Window managers
            # Not sure if this will work with e.g. Jupyter (does it know when the notebook is active?)
            if hasattr(self.canvas, "isActiveWindow"):
                if not self.canvas.isActiveWindow():
                    self.canvas.request_draw()
                    return
        elif rm == "reactive":
            # If the scene is not stale, we can skip rendering
            if not getattr(self, "_render_stale", False):
                self.canvas.request_draw()
                return

        # Now render the scene
        if self._show_fps:
            with self.stats:
                self.renderer.render(self.scene, self.camera, flush=False)
                if self.transform_gizmo:
                    self.renderer.render(self.transform_gizmo, self.camera, flush=False)
                self.renderer.render(
                    self.overlay_scene, self.overlay_camera, flush=False
                )
            self.stats.render()
        else:
            self.renderer.render(self.scene, self.camera, flush=False)
            if self.transform_gizmo:
                self.renderer.render(self.transform_gizmo, self.camera, flush=False)
            self.renderer.render(self.overlay_scene, self.overlay_camera)

        # Set stale to False
        self._render_stale = False

        self.canvas.request_draw()

    def _next_color(self):
        """Return next color in the colormap."""
        # Cache the full palette. N.B. that ordering of colors in cmap depends on
        # the number of colors requested - i.e. we can't just grab the last color.
        if not hasattr(self, "_cached_palette") or self.palette != self._cached_palette:
            self._cached_colors = list(cmap.Colormap(self.palette).iter_colors())
            self._cached_palette = self.palette

        if not hasattr(self, "_palette_index"):
            self._palette_index = -1
        self._palette_index += 1

        return self._cached_colors[self._palette_index % len(self._cached_colors)]

    def _next_label(self, prefix="Object"):
        """Return next label."""
        existing = [o for o in self.objects if str(o).startswith(prefix)]
        if len(existing) == 0:
            return prefix
        return f"{prefix}.{len(existing) + 1:03}"

    def __getitem__(self, key):
        """Get item."""
        return self.objects[key]

    def __contains__(self, key):
        """Check if object is on canvas."""
        return key in self.objects

    def __len__(self):
        """Return number of objects on canvas."""
        return len(self._object_ids)

    @property
    def blend_mode(self):
        """Deprecated! Render blend mode.

        This property has been deprecated. Please use `Viewer.set_alpha_mode()` instead.

        """
        raise DeprecationWarning(
            "The 'blend_mode' property is deprecated. Please use 'Viewer.set_alpha_mode()' instead."
        )

    @property
    def render_trigger(self):
        """Determines when the scene is (re)rendered.

        By default, we leave it to the renderer to decide when to render the scene.
        You can adjust that behaviour by setting render mode to:
         - "continuous" (default): leave it to the renderer to decide when to render the scene
         - "reactive": rendering is only triggered when the scene changes
         - "active_window": rendering is only done when the window is active; this currently
           only works with the PySide backend

        """
        return self._render_trigger

    @render_trigger.setter
    def render_trigger(self, mode):
        valid = ("continuous", "active_window", "reactive")
        if mode not in valid:
            raise ValueError(f"Unknown render mode: {mode}. Must be one of {valid}.")

        # No need to do anything if the value is the same
        if mode == getattr(self, "_render_trigger", None):
            return

        # Add/remove event handlers as necessary
        if mode == "reactive":
            self._set_stale_func = lambda event: setattr(self, "_render_stale", True)
            self.renderer.add_event_handler(
                self._set_stale_func,
                "pointer_down",
                "pointer_move",
                "pointer_up",
                "wheel",
                # "before_render",
            )
        elif self._render_trigger == "reactive":
            self.renderer.remove_event_handler(
                self._set_stale_func,
                "pointer_down",
                "pointer_move",
                "pointer_up",
                "wheel",
                # "before_render",
            )

        self._render_trigger = mode

    @property
    def controls(self):
        """Return the controls widget."""
        return getattr(self, "_controls", None)

    @property
    def visible(self):
        """List IDs of currently visible objects."""
        objects = self.objects  # grab this only once to speed things up
        return [s for s in objects if objects[s][0].visible]

    @property
    def invisible(self):
        """List IDs of currently visible objects."""
        objects = self.objects  # grab this only once to speed things up
        return [s for s in objects if not objects[s][0].visible]

    @property
    def pinned(self):
        """List IDs of currently pinned objects."""
        objects = self.objects  # grab this only once to speed things up
        return [s for s in objects if getattr(objects[s][0], "_pinned", False)]

    @property
    def selected(self):
        """Return IDs of or set selected objects."""
        return self._selected

    @selected.setter
    def selected(self, val):
        val = utils.make_iterable(val) if val is not None else []

        objects = self.objects  # grab once to speed things up
        logger.debug(f"{len(val)} objects selected ({len(self.selected)} previously)")
        # First un-highlight neurons which aren't selected anymore
        for s in [s for s in self._selected if s not in val]:
            for v in objects[s]:
                v.material.color = v._stored_color

        # Highlight new additions
        for s in val:
            if s not in self._selected:
                for v in objects[s]:
                    # Keep track of old colour
                    v._stored_color = v.material.color
                    v.material.color = gfx.Color(self.highlight_color)
        self._selected = list(val)

        # Update legend and set render stale (if applicable)
        update_helper(self, legend=True, bounds=False)

    @property
    def size(self):
        """Return size of the canvas."""
        return self.canvas.get_logical_size()

    @size.setter
    def size(self, size):
        """Set size of the canvas."""
        assert len(size) == 2
        self.canvas.set_logical_size(*size)

    @property
    def shadows(self):
        """Return shadow state."""
        return self._shadows

    @shadows.setter
    def shadows(self, v):
        """Set shadow state."""
        if not isinstance(v, bool):
            raise TypeError(f"Expected bool, got {type(v)}")

        def set_shadow(obj, state):
            if hasattr(obj, "cast_shadow"):
                obj.cast_shadow = state
            if hasattr(obj, "receive_shadow"):
                obj.receive_shadow = state

        if v != self._shadows:
            self._shadows = v
            for vis in self.visuals:
                set_shadow(vis, v)

            for ch in self.scene.children:
                if isinstance(ch, gfx.PointLight):
                    ch.cast_shadow = v

            # self.scene.traverse(lambda x: set_shadow(x, v))

    @property
    def visuals(self):
        """List of all visuals on this canvas."""
        return [c for c in self.scene.children if hasattr(c, "_object_id")]

    @property
    def bounds(self):
        """Bounds of all currently visuals (visible and invisible)."""
        bounds = []
        for vis in self.visuals:
            # Skip the bounding box itself
            if getattr(vis, "_object_id", "") == "boundingbox":
                continue

            try:
                bounds.append(vis._bounds)
            except BaseException:
                pass

        if not bounds:
            return None

        bounds = np.dstack(bounds)

        mn = bounds[:, 0, :].min(axis=1)
        mx = bounds[:, 1, :].max(axis=1)

        return np.vstack((mn, mx)).T

    @property
    def max_fps(self):
        """Maximum frames per second to render."""
        return self.canvas._subwidget._BaseRenderCanvas__scheduler._max_fps

    @max_fps.setter
    def max_fps(self, v):
        assert isinstance(v, int)
        self.canvas._subwidget._BaseRenderCanvas__scheduler._max_fps = v

    @property
    def moveable_object(self):
        """Get/Set the object that can be moved via the transform gizmo."""
        if self.transform_gizmo is None:
            return None
        return self.transform_gizmo._object_to_control

    @moveable_object.setter
    def moveable_object(self, obj):
        if obj is None:
            if self.transform_gizmo:
                self.transform_gizmo._object_to_control = None
            return

        if isinstance(obj, str):
            if obj not in self.objects:
                raise ValueError(f"Object '{obj}' not found.")
            elif len(self.objects[obj]) > 1:
                raise ValueError(f"Object '{obj}' consists of multiple WorldObjects.")
            obj = self.objects[obj][0]
        elif not isinstance(obj, gfx.WorldObject):
            raise TypeError(f"Expected pygfx object, got {type(obj)}")

        if self.transform_gizmo is None:
            # The transform gizmo is rendered independent of the scene (so it always stay on top)
            self.transform_gizmo = gfx.TransformGizmo(obj)
            self.transform_gizmo.add_default_event_handlers(self.renderer, self.camera)
        else:
            self.transform_gizmo._object_to_control = obj

    @property
    def _is_jupyter(self):
        """Check if Viewer is using Jupyter canvas."""
        return "Jupyter" in str(type(self.canvas))

    @property
    def _is_offscreen(self):
        """Check if Viewer is using offscreen canvas."""
        return isinstance(self.canvas, OffscreenRenderCanvas)

    @property
    def _window_manager(self):
        """Which window manager is being used."""
        try:
            return type(self.canvas).__module__.split(".")[-1]
        except BaseException:
            return "na"

    @property
    def _object_ids(self):
        """All object IDs on this canvas in order of addition."""
        obj_ids = []
        for v in self.visuals:
            if hasattr(v, "_object_id"):
                obj_ids.append(v._object_id)
        return sorted(set(obj_ids), key=lambda x: obj_ids.index(x))

    @property
    def objects(self):
        return self._objects()

    @lru_cache(maxsize=1)
    def _objects(self):
        """Ordered dictionary {name->[visuals]} of all objects in order of addition."""
        objects = OrderedDict()
        for v in self.visuals:
            if hasattr(v, "_object_id"):
                if v._object_id in objects:
                    objects[v._object_id].append(v)
                else:
                    objects[v._object_id] = [v]

        return objects

    @property
    def objects_pickable(self):
        return self._objects_pickable

    @objects_pickable.setter
    def objects_pickable(self, v):
        if not isinstance(v, bool):
            raise TypeError(f"Expected bool, got {type(v)}")

        # No need to do anything if the value is the same
        if v == self._objects_pickable:
            return

        self._objects_pickable = v

        # Set pick_write to new value for all materials
        for objects in self.objects.values():
            for ob in objects:
                try:
                    ob.material.pick_write = v
                except AttributeError:
                    pass

    @property
    def highlighted(self):
        """Return IDs of currently highlighted objects."""
        highlighted = []
        for obj in self.objects:
            if any([getattr(v, "_highlighted", False) for v in self.objects[obj]]):
                highlighted.append(obj)
        return highlighted

    @property
    def on_hover(self):
        """Determines what to do when hovering over objects.

        Can be set to:
         - `None`: do nothing
         - "highlight": hide object

        """
        return self._on_hover

    @on_hover.setter
    def on_hover(self, v):
        valid = (None, "highlight")
        if v not in valid:
            raise ValueError(
                f"Unknown value for on_hover: {v}. Must be one of {valid}."
            )

        # No need to do anything if the value is the same
        if v == self._on_hover:
            return

        if v:
            # Make objects pickable
            self.objects_pickable = True

            # Add the event handler
            self.scene.add_event_handler(self._highlight_on_hover_event, "pointer_move")
        else:
            self.scene.remove_event_handler(
                self._highlight_on_hover_event, "pointer_move"
            )
            current_hover = getattr(self, "_current_hover_object", None)

            # Make sure to unhighlight the current hover object
            if current_hover:
                self.unhighlight_objects(current_hover)
                self._current_hover_object = None

        self._on_hover = v

    def _highlight_on_hover_event(self, event):
        """This is the event callback for highlighting objects on hover."""
        # If any buttons are pressed (e.g. mouse left during panning) ignore the event
        if event.buttons:
            return

        # Parse the current object
        new_hover = event.pick_info["world_object"]
        current_hover = getattr(self, "_current_hover_object", None)

        # Break early if there is nothing to do
        if new_hover is None and current_hover is None:
            # print("  No hover")
            return

        new_hover_id = [k for k, v in self.objects.items() if new_hover in v]
        new_hover_id = new_hover_id[0] if new_hover_id else None

        # See if we need to de-highlight the current hover object
        if current_hover:
            # If the new object is the same as the current one, we don't need to do anything
            if current_hover == new_hover_id:
                return
            if current_hover in self.objects:
                self.unhighlight_objects(current_hover)
            self._current_hover_object = None

        # Highlight the new object
        if new_hover_id:
            self.highlight_objects(
                new_hover_id, color=getattr(self, "_highlight_on_hover_color", 0.2)
            )
            self._current_hover_object = new_hover_id

    @property
    def on_double_click(self):
        """Determines what to do when double clicking on objects.

        Can be set to:
         - `None`: do nothing
         - "hide": hide object
         - "remove": remove object
         - "select": select object

        """
        return self._on_double_click

    @on_double_click.setter
    def on_double_click(self, v):
        valid = (None, "hide", "remove", "select")
        if v not in valid:
            raise ValueError(
                f"Unknown value for on_double_click: {v}. Must be one of {valid}."
            )

        # No need to do anything if the value is the same
        if v == self._on_double_click:
            return

        # First try to remove the current event handler for double clicks
        try:
            self.scene.remove_event_handler(
                getattr(self, "_on_double_click_func", None), "double_click"
            )
        except KeyError:
            pass

        if v:
            # Make objects pickable
            self.objects_pickable = True

            # Now add the new event handler
            func = partial(handle_object_event, viewer=self, actions=(v,))
            self.scene.add_event_handler(func, "double_click")
            self.__on_double_click_func = func

        self._on_double_click = v

    def add_animation(self, x, on_error="remove", run_every=None, req_render=True):
        """Add animation function to the Viewer.

        Parameters
        ----------
        x :         callable
                    Function to add to the animation loop.
        on_error :  "remove" | "ignore" | "raise" | "log"
                    What to do if the function throws an error. If "remove",
                    the function will be removed from the animation loop. If
                    "ignore", the error will be ignored and the function will
                    continue to be called.
        run_every : int, optional
                    Use to run the function every n frames.
        req_render : bool, optional
                    Whether this animation requires a re-render of the scene.
                    This is mainly a flag to help the viewer to decide
                    whether/when to trigger a render. See also the `render_trigger`
                    property.

        """
        if not callable(x):
            raise TypeError(f"Expected callable, got {type(x)}")

        assert on_error in ["remove", "ignore", "raise", "log"]

        self._animations[x] = (on_error, run_every, req_render)

    def remove_animation(self, x):
        """Remove animation function from the Viewer.

        Parameters
        ----------
        x :     callable | int
                Either the function itself or its index
                in the list of animations.

        """
        if callable(x):
            self._animations_flagged_for_removal.append(x)
        elif isinstance(x, int):
            self._animations_flagged_for_removal.append(
                list(self._animations.keys())[x]
            )
        else:
            raise TypeError(f"Expected callable or index (int), got {type(x)}")

    def add_effect(self, effect, **kwargs):
        """Add post-processing effect to the renderer.

        You can also use this method to adjust the parameters of an existing effect.

        Parameters
        ----------
        effect :   str
                    Name of the effect to add. Currently supported:
                     - "edl" (Eye-Dome Lighting)
                       This effect enhances depth perception for complex
                       geometries by darkening edges based on depth differences.

        **kwargs
                    Keyword arguments passed to the effect constructor:
                    - edl:
                      - strength (default 5): EDL strength; ypical range ~ [0.5, 10.0].
                      - radius (default 1.5): sampling radius in pixels
                      - depth_edge_threshold (default 0.0)

        """
        effect = EFFECT_CLASSES.get(effect, None)
        if effect is None:
            raise ValueError(f"Unknown effect: {effect}")

        # Check if we already have this effect
        p = None
        for e in self.renderer.effect_passes:
            if isinstance(e, effect):
                p = e
                break

        if p is None:
            # Overwrite the default of 1 (seems too weak in my hands)
            if (effect is EDLPass) and "strength" not in kwargs:
                kwargs["strength"] = 5.0

            p = effect(**kwargs)
            self.renderer.effect_passes = tuple(list(self.renderer.effect_passes) + [p])
        else:
            # Update parameters
            for k, v in kwargs.items():
                if hasattr(p, k):
                    setattr(p, k, v)
                else:
                    raise ValueError(f"Effect '{effect}' has no parameter '{k}'")

    def show(self, use_sidecar=False, toolbar=False, start_loop=False):
        """Show viewer.

        Parameters
        ----------

        For Jupyter lab only:

        use_sidecar : bool
                      If True, will use the Sidecar extension to display the
                      viewer outside the notebooks. Will throw an error if
                      Sidecar is not installed.
        toolbar :     bool
                      If True, will show a toolbar. You can always show/hide
                      the toolbar with ``viewer.show_controls()`` and
                      ``viewer.hide_controls()``, or the `c` hotkey.

        For scripts & standard REPL:

        start_loop :  bool
                      If True, will start the blocking (!) event loop. This is
                      the recommended way to show the viewer when using it in a script.
                      From an interactive REPL such as IPython you should be able to
                      just call ``Viewer.show()`` and the interactive viewer will appear
                      while still allowing you to interact with the REPL.

        """
        # This is for e.g. headless testing
        if getattr(config, "HEADLESS", False):
            logger.info("Viewer widget not shown - running in headless mode.")
            return

        # Start the animation loop
        self.canvas.request_draw(self._animate)

        # If this is an offscreen canvas, we don't need to do anything else
        if isinstance(self.canvas, OffscreenRenderCanvas):
            return

        # In terminal we can just show the window
        if not self._is_jupyter:
            # Not all backends have a show method (e.g. GLFW does not)
            if hasattr(self.canvas, "show"):
                self.canvas.show()

            if start_loop:
                from rendercanvas.auto import loop

                loop.run()
            elif utils._type_of_script() in ("terminal", "script") and os.environ.get(
                "OCTARINE_CHECK_LOOP", "1"
            ) in ("1", "true", "True"):
                logger.warning(
                    "Running in a (potentially) non-interactive terminal or script "
                    "environment. You may have to manually start the event loop "
                    "for the canvas to render:\n\n"
                    "  >>> v = octarine.Viewer(show=False)\n"
                    "  >>> ...  # setup your viewer\n"
                    "  >>> v.show(start_loop=True)\n\n"
                    "Alternatively, use the loop.run() function:\n\n"
                    "  >>> from rendercanvas.auto import loop\n"
                    "  >>> ...  # setup your viewer\n"
                    "  >>> v.show()\n"
                    "  >>> loop.run()\n\n"  # do not remove the \n\n here
                )
        else:
            # if not hasattr(self, 'widget'):
            from .jupyter import JupyterOutput
            from IPython.display import display

            # Construct the widget
            if not hasattr(self, "widget"):
                self.widget = JupyterOutput(
                    self,
                    use_sidecar=use_sidecar,
                    toolbar=toolbar,
                    sidecar_kwargs={"title": self._title},
                )

            # This will display the viewer right here and there
            display(self.widget)

    def show_message(
        self, message, position="top-right", font_size=20, color=None, duration=None
    ):
        """Show message on canvas.

        Parameters
        ----------
        message :   str | None
                    Message to show. Set to `None` to remove the existing message.
        position :  "top-left" | "top-right" | "bottom-left" | "bottom-right" | "center"
                    Position of the message on the canvas.
        font_size : int, optional
                    Font size of the message.
        color :     str | tuple, optional
                    Color of the message. If `None`, will use white.
        duration :  int, optional
                    Number of seconds after which to fade the message.

        """
        if message is None and hasattr(self, "_message_text"):
            if self._message_text.parent:
                self.overlay_scene.remove(self._message_text)
            del self._message_text
            return

        _positions = {
            "top-left": (-0.95, 0.95, 0),
            "top-right": (0.95, 0.95, 0),
            "bottom-left": (-0.95, -0.95, 0),
            "bottom-right": (0.95, -0.95, 0),
            "center": (0, 0, 0),
        }
        if position not in _positions:
            raise ValueError(f"Unknown position: {position}")

        if not hasattr(self, "_message_text"):
            self._message_text = text2gfx(
                message, color="white", font_size=font_size, screen_space=True
            )

        # Make sure the text is in the scene
        if self._message_text not in self.overlay_scene.children:
            self.overlay_scene.add(self._message_text)

        self._message_text.set_text(message)
        self._message_text.font_size = font_size
        self._message_text.anchor = position
        if color is not None:
            self._message_text.material.color = cmap.Color(color).rgba
        self._message_text.material.opacity = 1
        self._message_text.local.position = _positions[position]

        # When do we need to start fading out?
        if duration:
            self._fade_out_time = time.time() + duration

            def _fade_message():
                if not hasattr(self, "_message_text"):
                    self.remove_animation(_fade_message)
                else:
                    if time.time() > self._fade_out_time:
                        # This means the text will fade fade over 1/0.02 = 50 frames
                        self._message_text.material.opacity = max(
                            self._message_text.material.opacity - 0.02, 0
                        )

                    if self._message_text.material.opacity <= 0:
                        if self._message_text.parent:
                            self.overlay_scene.remove(self._message_text)
                        self.remove_animation(_fade_message)

            self.add_animation(_fade_message)

    def show_controls(self):
        """Show controls."""
        if self._is_jupyter:
            if self.widget.toolbar:
                self.widget.toolbar.show()
        else:
            if not hasattr(self, "_controls"):
                from .controls import Controls

                self._controls = Controls(self)
            self._controls.show()

    def hide_controls(self):
        """Hide controls."""
        if self._is_jupyter:
            if self.widget.toolbar:
                self.widget.toolbar.hide()
        else:
            if hasattr(self, "_controls"):
                self._controls.hide()

    def _toggle_controls(self):
        """Switch controls on and off."""
        if self._is_jupyter:
            if self.widget.toolbar:
                self.widget.toolbar.toggle()
        else:
            if not hasattr(self, "_controls"):
                self.show_controls()
            elif self._controls.isVisible():
                self.hide_controls()
            else:
                self.show_controls()

    @update_viewer(legend=True, bounds=True)
    def clear(self):
        """Clear canvas of objects (expects lights and background)."""
        # Skip if running in headless mode
        if getattr(config, "HEADLESS", False):
            return

        # Remove everything but the lights and backgrounds
        self.scene.remove(*self.visuals)

        # Rset the transform gizmo
        self.transform_gizmo = None

    @update_viewer(legend=True, bounds=True)
    def remove_objects(self, to_remove):
        """Remove given neurons/visuals from canvas."""
        to_remove = utils.make_iterable(to_remove)

        for vis in self.scene.children:
            if vis in to_remove:
                self.scene.remove(vis)
            elif hasattr(vis, "_object_id"):
                if vis._object_id in to_remove:
                    self.scene.remove(vis)

    @update_viewer(legend=True, bounds=True)
    def pop(self, N=1):
        """Remove the most recently added N visuals."""
        for vis in list(self.objects.values())[-N:]:
            self.remove_objects(vis)

    @property
    def show_bounds(self):
        """Set to ``True`` to show bounding box."""
        return self._show_bounds

    @property
    def show_fps(self):
        """Show frames per second."""
        return self._show_fps

    @show_fps.setter
    def show_fps(self, v):
        if not isinstance(v, bool):
            raise TypeError(f"Expected bool, got {type(v)}")
        self._show_fps = v
        self._render_stale = True

    def toggle_bounds(self):
        """Toggle bounding box."""
        self.show_bounds = not self.show_bounds

    @show_bounds.setter
    def show_bounds(self, v):
        if not isinstance(v, bool):
            raise TypeError(f"Need bool, got {type(v)}")

        self._show_bounds = v

        if self.show_bounds:
            self.update_bounds()
        else:
            self.remove_bounds()

    def remove_bounds(self):
        """Remove bounding box visual."""
        self._show_bounds = False
        for v in self.visuals:
            if getattr(v, "_object_type", "") == "boundingbox":
                self.remove_objects(v)

    def resize(self, size):
        """Resize canvas.

        Parameters
        ----------
        size :  (width, height) tuple
                New size of the canvas.
        """
        assert len(size) == 2
        self.canvas.set_logical_size(*size)

    def update_bounds(self, color="w", width=1):
        """Update bounding box visual."""
        # Remove any existing visual
        self.remove_bounds()

        self._show_bounds = True

        # Skip if no visual on canvas
        bounds = self.scene.get_bounding_box()
        if isinstance(bounds, type(None)):
            return

        # Create box visual
        box = gfx.BoxHelper()
        box.set_transform_by_aabb(bounds)

        # Add custom attributes
        box._object_type = "boundingbox"
        box._object_id = uuid.uuid4()

        self.scene.add(box)

    def center_camera(self):
        """Center camera on visuals."""
        if len(self):
            self.camera.show_object(
                self.scene, scale=1, view_dir=(0.0, 0.0, 1.0), up=(0.0, -1.0, 0.0)
            )

    @update_viewer(legend=True, bounds=True)
    def add(self, x, name=None, group=None, center=True, clear=False, **kwargs):
        """Add object to canvas.

        This function is a general entry point for adding objects to the canvas.
        It will look at the type of the input and try to find an appropriate
        function to convert the input to visuals.

        Use `octarine.register_converter` to add custom converters.

        Parameters
        ----------
        x
                    Object(s) to add to the canvas.
        name :      str, optional
                    Name for the visual(s).
        group :     str, optional
                    Group for the visual(s).
        center :    bool, optional
                    If True, re-center camera to all objects on canvas.
        clear :     bool, optional
                    If True, clear canvas before adding new objects.
        **kwargs
                    Keyword arguments passed to the conversion functions when
                    generating visuals.

        Returns
        -------
        None

        """
        if clear:
            self.clear()

        converter = get_converter(x, raise_missing=False)
        if utils.is_iterable(x) and not converter:
            for xx in x:
                self.add(xx, center=False, clear=False, name=name, **kwargs)
            if center:
                self.center_camera()
            return

        if converter is None:
            raise NotImplementedError(f"No converter found for {x} ({type(x)})")

        # Check if we have to provide a color
        if "color" not in kwargs and "color" in inspect.signature(converter).parameters:
            kwargs["color"] = tuple(self._next_color().rgba)

        visuals = utils.make_iterable(converter(x, **kwargs))

        for v in visuals:
            # If we have a name, assign it to the visual
            if name is not None:
                v._object_id = name
            # If not we either use existing ID or generate a new one
            else:
                # Give visuals an _object_id if they don't already have one
                if not hasattr(v, "_object_id"):
                    new_id = self._next_label("Object")
                    for v2 in visuals:
                        v._object_id = new_id
                elif not isinstance(v._object_id, str):
                    v._object_id = str(v._object_id)

            v._object_group = group

            self.scene.add(v)

        if center:
            self.center_camera()

    @update_viewer(legend=True, bounds=True)
    def _add_to_scene(self, visual, center=True):
        """Add visual to scene.

        This is just a convenient collection point for us to trigger a bunch of updates in one go,
        """
        # If we need objects to be pickable, set the material accordingly
        if self.objects_pickable:
            try:
                visual.material.pick_write = True
            except AttributeError:
                pass

        self.scene.add(visual)

        if center:
            self.center_camera()

    def add_mesh(
        self, mesh, name=None, group=None, color=None, alpha=None, center=True
    ):
        """Add mesh to canvas.

        Parameters
        ----------
        mesh :      Mesh-like
                    Mesh to plot. If this is a pygfx.Mesh, it will be added
                    directly to the scene without modification (i.e. `color`,
                    `alpha`, etc. will be ignored).
        name :      str, optional
                    Name for the visual.
        group :     str, optional
                    Group for the visual.
        color :     str | tuple, optional
                    Color to use for plotting. If multiple colors,
                    must be a list of colors with the same length as
                    the number of faces or vertices.
        alpha :     float, optional
                    Opacity value [0-1]. If provided, will override
                    the alpha channel of the color.
        center :    bool, optional
                    If True, re-center camera to all objects on canvas.

        """
        if isinstance(mesh, tm.Scene):
            for _, ob in mesh.geometry.items():
                self.add_mesh(ob, name=name, color=color, alpha=alpha, center=False)
            return

        if not utils.is_mesh_like(mesh):
            raise TypeError(f"Expected mesh-like object, got {type(mesh)}")
        if color is None:
            color = self._next_color()
        if name is None:
            name = self._next_label("Mesh")
        elif not isinstance(name, str):
            name = str(name)

        if not isinstance(mesh, gfx.Mesh):
            visual = mesh2gfx(mesh, color=color, alpha=alpha)
        else:
            visual = mesh

        visual._object_id = name if name else uuid.uuid4()
        visual._object_group = group

        self._add_to_scene(visual, center)

    def add_points(
        self,
        points,
        name=None,
        group=None,
        color=None,
        marker=None,
        size=2,
        size_space="screen",
        center=True,
    ):
        """Add points plot to canvas.

        Parameters
        ----------
        points :    (N, 3) array
                    Points to plot.
        name :      str, optional
                    Name for the visual.
        group :     str, optional
                    Group for the visual.
        color :     str | tuple, optional
                    Color to use for plotting. Can be the name of
                    a colormap or a single color.
        marker :    str, optional
                    Marker to use for plotting. By default (None), will
                    use a point. Other options include e.g. "circle", "ring"
                    or "diamond". See `pygfx.MarkerShape` for the definitive
                    list of options. Please note that you may have to
                    increase the size of the marker to see some of the shapes.
        size :      int | float
                    Marker size. Can be a single value or an array of
                    sizes for each point.
        size_space : "screen" | "world" | "model", optional
                    Units to use for the marker size. "screen" (default)
                    will keep the line width constant on the screen, while
                    "world" and "model" will keep it constant in world and
                    model coordinates, respectively. In the latter two cases,
                    `size` corresponds to the diameter (not radius) of the
                    marker!
        center :    bool, optional
                    If True, re-center camera to all objects on canvas.

        """
        if not isinstance(points, np.ndarray):
            raise TypeError(f"Expected numpy array, got {type(points)}")
        if points.ndim != 2 or points.shape[1] != 3:
            raise ValueError(f"Expected (N, 3) array, got {points.shape}")
        if color is None:
            color = self._next_color()
        if name is None:
            name = self._next_label("Scatter")
        elif not isinstance(name, str):
            name = str(name)

        visual = points2gfx(
            points, color=color, size=size, size_space=size_space, marker=marker
        )
        visual._object_id = name if name else uuid.uuid4()
        visual._object_group = group
        self._add_to_scene(visual, center)

    def add_lines(
        self,
        lines,
        name=None,
        group=None,
        color=None,
        linewidth=1,
        linewidth_space="screen",
        linestyle="solid",
        center=True,
    ):
        """Add lines to canvas.

        Parameters
        ----------
        lines :     list of (N, 3) arrays | (N, 3) array
                    Lines to plot. If a list of arrays, each array
                    represents a separate line. If a single array,
                    each row represents a point in the line. You can
                    introduce breaks in the line by inserting NaNs.
        name :      str, optional
                    Name for the visual.
        group :     str, optional
                    Group for the visual.
        color :     str | tuple, optional
                    Color to use for plotting. Can be a single color
                    or one for every point in the line(s).
        linewidth : float, optional
                    Line width.
        linewidth_space : "screen" | "world" | "model", optional
                    Units to use for the line width. "screen" (default)
                    will keep the line width constant on the screen, while
                    "world" and "model" will keep it constant in world and
                    model coordinates, respectively.
        linestyle : "solid" | "dashed" | "dotted" | "dashdot" | tuple, optional
                    Line style to use. If a tuple, must define the on/off
                    sequence.
        center :    bool, optional
                    If True, re-center camera to all objects on canvas.

        """
        # TODO:
        # - allow providing a tuple of (positions, edges) for lines

        if isinstance(lines, np.ndarray):
            if lines.ndim != 2 or lines.shape[1] != 3:
                raise ValueError(f"Expected (N, 3) array, got {lines.shape}")
        elif isinstance(lines, list):
            if not all([l.ndim == 2 and l.shape[1] == 3 for l in lines]):
                raise ValueError("Expected list of (N, 3) arrays.")
        else:
            raise TypeError(f"Expected numpy array or list, got {type(lines)}")

        if color is None:
            color = self._next_color()
        if name is None:
            name = self._next_label("Lines")
        elif not isinstance(name, str):
            name = str(name)

        visual = lines2gfx(
            lines,
            linewidth=linewidth,
            linewidth_space=linewidth_space,
            color=color,
            dash_pattern=linestyle,
        )
        visual._object_id = name if name else uuid.uuid4()
        visual._object_group = group
        self._add_to_scene(visual, center)

    def add_volume(
        self,
        volume,
        spacing=(1, 1, 1),
        name=None,
        group=None,
        color=None,
        opacity=1.0,
        offset=(0, 0, 0),
        clim="data",
        slice=False,
        interpolation="linear",
        hide_zero=True,
        center=True,
    ):
        """Add image volume to canvas.

        Note that the default blend mode for the renderer may cause objects
        behind or inside the volume to look funny. You can change the blend
        mode by setting e.g. `viewer.blend_mode='additive'`.

        Parameters
        ----------
        volume :    (N, M, K) array
                    Volume to plot.
        spacing :   tuple
                    Spacing between voxels.
        name :      str, optional
                    Name for the visual.
        group :     str, optional
                    Group for the visual.
        color :     color | list of colors | pygfx.Texture, optional
                    Colormap to render the volume. This can be:
                      - name of a colormap (e.g. "viridis" or "magma")
                      - a single color (name, hex, rgb, rgba)
                      - a list of colors
                      - a 1D pygfx.Texture
                    Note that single colors typically don't look good and
                    it's better to define at least two colors. For example,
                    instead of "red" use ["red", "yellow"]. If `None` will
                    use one of the built-in pygfx colormaps.
        opacity :   float, optional
                    Overall opacity of the volume. Must be between 0 and 1.
        offset :    tuple, optional
                    (x, y, z) offset for the volume. If None, will use (0, 0, 0).
        clim :      "data" | "datatype" | tuple, optional
                    The contrast limits to scale the data values with.
                      - "data" (default) will use the min/max of the data
                      - "datatype" will use (0, theoretical max of data type)
                        for integer data, e.g. (0, 255) for int8 and uint8,
                        and (0, 1) for float data assuming the data has been
                        normalized
                      - tuple of min/max values or combination of "data" and
                        "datatype" strings
        slice :         bool | tuple, optional
                        Render volume slices instead of the full volume:
                        - True: render slices along all three dimensions
                        - tuple of bools, e.g. `(True, True, False)`: render slices
                          in the respective dimensions
                        - tuple of floats, e.g. `(0.5, 0.5, 0.5)`: render slices
                          at the respective positions (relative to the volume size)
        interpolation : "linear" | "nearest"
                    Interpolation to use when rendering the volume. "linear"
                    (default) looks better but is slower.
        hide_zero : bool
                    If True, will hide voxels with lowest value according to `cmin`.
        center :    bool, optional
                    If True, re-center camera to all objects on canvas.

        """
        if not isinstance(volume, np.ndarray):
            raise TypeError(f"Expected numpy array, got {type(volume)}")
        if volume.ndim != 3:
            raise ValueError(f"Expected 3D array, got {volume.ndim}")
        if name is None:
            name = self._next_label("Volume")
        elif not isinstance(name, str):
            name = str(name)

        visuals = volume2gfx(
            volume,
            spacing=spacing,
            offset=offset,
            color=color,
            opacity=opacity,
            clim=clim,
            slice=slice,
            interpolation=interpolation,
            hide_zero=hide_zero,
        )
        name = name if name else uuid.uuid4()
        for vis in visuals:
            vis._object_id = name if name else uuid.uuid4()
            vis._object_group = group
            self._add_to_scene(vis, center)

    def close(self):
        """Close the viewer."""
        # Skip if this is headless mode
        if getattr(config, "HEADLESS", False):
            return

        # Clear first to free all visuals
        self.clear()

        # Remove from config if this is the primary viewer
        if self == getattr(config, "PRIMARY_VIEWER", None):
            del config.PRIMARY_VIEWER

        # Close if not already closed
        if not self.canvas.get_closed():
            self.canvas.close()

        if hasattr(self, "_controls"):
            self._controls.close()

        # Close the Jupyter widget
        if hasattr(self, "widget") and not getattr(self.widget, "_is_closed", False):
            self.widget.close(close_viewer=False)

        try:
            viewers.remove(self)
        except ValueError:
            pass

    @update_viewer(legend=True, bounds=True)
    def hide_objects(self, obj):
        """Hide given object(s).

        Parameters
        ----------
        obj :   str | list
                Object(s) to hide.

        """
        objects = self.objects  # grab once to speed things up
        for ob in utils.make_iterable(obj):
            if ob not in objects:
                logger.warning(f'Object "{ob}" not found on canvas.')
                continue
            for v in objects[ob]:
                if getattr(v, "_pinned", False):
                    continue
                if v.visible:
                    v.visible = False

    @update_viewer(legend=True, bounds=True)
    def hide_selected(self):
        """Hide currently selected object(s)."""
        self.hide_neurons(self.selected)

    @update_viewer(legend=True, bounds=True)
    def unhide_objects(self, obj=None):
        """Unhide given object(s).

        Parameters
        ----------
        obj :   str | list | None
                Object(s) to unhide. If None, will unhide all objects.

        """
        objects = self.objects  # grab once to speed things up
        if obj is not None:
            ids = utils.make_iterable(obj)
        else:
            ids = list(objects.keys())

        for ob in ids:
            if ob not in objects:
                logger.warning(f"Object {ob} not found on canvas.")
                continue
            for v in objects[ob]:
                if getattr(v, "_pinned", False):
                    continue
                if not v.visible:
                    v.visible = True

    def highlight_objects(self, obj, color=0.2):
        """Highlight given object(s) by increasing their brightness.

        Parameters
        ----------
        obj :   str | int | list | visual
                Object(s) to highlight. Can be the name(s) or ID(s) of
                the object(s), their index(es) in the list of visuals,
                or the visual(s) themselves. Objects already highlighted
                will be silently ignored.
        color : float | tuple
                Color to use for highlighting. If a float, will change
                the HSV value of the current color. If a tuple, will
                use the RGB(A) color.

        See Also
        --------
        Viewer.unhighlight_objects
                Use to remove highlights.

        """
        if not utils.is_iterable(obj):
            objects = [obj]
        else:
            objects = obj

        all_objects = self.objects  # grab once to speed things up

        for ob in objects:
            if ob in all_objects:
                list_ = all_objects[ob]
            elif isinstance(ob, int):
                list_ = list(self.objects.values())[ob]
            elif isinstance(ob, gfx.WorldObject):
                list_ = [ob]
            else:
                raise TypeError(f"Unknown object type: {type(ob)}")

            for o in list_:
                # Skip if object is pinned
                if getattr(o, "_pinned", False):
                    continue
                # Skip if object is already highlighted
                if getattr(o, "_highlighted", False):
                    continue

                if isinstance(color, (float, int)):
                    # Work in HSL space
                    h, s, l = o.material.color.to_hsl()
                    # If the value is not maxed yet, increase it
                    if l < 1:
                        l = min(l + color, 1)
                    else:
                        l = max(l - color, 0)

                    new_color = gfx.Color.from_hsl(h, s, l)
                else:
                    # See if pygfx can handle the color
                    new_color = gfx.Color(color)

                o.material._original_color = o.material.color
                o.material.color = new_color
                o._highlighted = True

    def unhighlight_objects(self, obj=None):
        """Unhighlight given object(s).

        Parameters
        ----------
        obj :   str | int | list | visual
                Object(s) to unhighlight. Can be the name(s) or ID(s) of
                the object(s), their index(es) in the list of visuals,
                or the visual(s) themselves. If None, will unhighlight all
                objects. Objects that aren't highlighted will be silently
                ignored.

        See Also
        --------
        Viewer.highlight_objects
                Use to highlight objects

        """
        # Important note: it looks like any attribute we added previously
        # will (at some point) have been silently renamed to "_Viewer{attribute}"
        if obj is None:
            obj = [v for v in self.visuals if getattr(v, "_highlighted", False)]

        if not utils.is_iterable(obj):
            objects = [obj]
        else:
            objects = obj

        all_objects = self.objects  # grab once to speed things up

        for ob in objects:
            if ob in all_objects:
                list_ = all_objects[ob]
            elif isinstance(ob, int):
                list_ = list(self.visuals.values())[ob]
            elif isinstance(ob, gfx.WorldObject):
                list_ = [ob]
            else:
                raise TypeError(f"Unknown object type: {type(ob)}")

            for o in list_:
                # Skip if object is pinned
                if getattr(o, "_pinned", False):
                    continue

                # Skip if object isn't actually highlighed
                if not getattr(o, "_highlighted", False):
                    continue
                o.material.color = o.material._original_color
                del o.material._original_color
                del o._highlighted

    def pin_objects(self, obj):
        """Pin given object(s).

        Changes to the color or visibility of pinned neurons are silently
        ignored. You can use this to keep specific neurons visible while
        cycling through the rest - useful for comparisons.

        """
        obj = utils.make_iterable(obj)
        objects = self.objects  # grab only once to speed things up

        for ob in obj:
            if ob not in objects:
                logger.warning(f"Object {ob} not found on canvas.")
                continue
            for v in objects[ob]:
                v._pinned = True

    def unpin_objects(self, obj=None):
        """Unpin given object(s).

        Use ``obj`` to unhide specific neurons.

        """
        objects = self.objects  # grab once to speed things up
        if obj is None:
            obj = objects
        else:
            obj = utils.make_iterable(obj)

        for ob in obj:
            if ob not in objects:
                logger.warning(f"Object {ob} not found on canvas.")
                continue
            for v in objects[ob]:
                v._pinned = False

    @update_viewer(legend=False, bounds=False)
    def set_alpha_mode(self, mode, objects=None):
        """Defines how objects' colors are blended.

        With version v0.13.0 pygfx replaced the single renderer.blend_mode property with
        customizable per-material alpha modes. The Viewer.set_alpha_mode function provides
        a high-level interface to these settings. If you need more fine-grained control,
        see the material.alpha_mode property of individual objects.

        Parameters
        ----------
        mode :      str
                    The mode to set. Please see the pygfx documentation for details:
                      >>> import pygfx
                      >>> help(pygfx.Material.alpha_mode)
        objects :   list, optional
                    Objects to set the alpha mode for. If None, will set for all objects.

        """
        if objects is None:
            objects = list(self.objects)

        for n in objects:
            for v in self.objects[n]:
                if getattr(v, "_pinned", False):
                    continue
                if not hasattr(v, "material"):
                    continue
                v.material.alpha_mode = mode

    @update_viewer(legend=True, bounds=False)
    def set_colors(self, c, alpha_mode="auto"):
        """Set object color.

        Parameters
        ----------
        c :     tuple | dict
                RGB color(s) to apply. Values must be 0-1. Accepted:
                   1. Tuple of single color. Applied to all visible objects.
                   2. Dictionary names/IDs to colors.
        alpha_mode : str
                If "auto" (default), will set the alpha mode to "add" if the
                opacity is < 1, and "opaque" otherwise. Set `alpha_mode` to `None` to
                skip this adjustment.

        """
        objects = self.objects  # grab once to speed things up
        if isinstance(c, (tuple, list, np.ndarray, str)):
            cmap = {s: c for s in objects}
        elif isinstance(c, dict):
            cmap = c
        else:
            raise TypeError(f'Unable to use colors of type "{type(c)}"')

        for n in objects:
            if n in cmap:
                for v in objects[n]:
                    if getattr(v, "_pinned", False):
                        continue
                    if not hasattr(v, "material"):
                        continue
                    # Note: there is currently a bug where removing or adding an alpha
                    # channel from a color will break the rendering pipeline
                    if len(v.material.color) == 4:
                        new_c = gfx.Color(cmap[n]).rgba
                    else:
                        new_c = gfx.Color(cmap[n]).rgb
                    v.material.color = gfx.Color(new_c)

                    # Determine if we consider this transparent
                    if len(new_c) == 4 and new_c[3] < 1:
                        is_transparent = True
                    elif v.material.opacity < 1:
                        is_transparent = True
                    else:
                        is_transparent = False

                    if alpha_mode == "auto":
                        if is_transparent:
                            v.material.alpha_mode = "add"
                        else:
                            v.material.alpha_mode = "solid"
                    elif alpha_mode:
                        v.material.alpha_mode = alpha_mode

    def colorize(self, palette="seaborn:tab10", objects=None, randomize=True):
        """Colorize objects using a color palette.

        Parameters
        ----------
        palette :   str | cmap Colormap
                    Name of the `cmap` palette to use. See
                    https://cmap-docs.readthedocs.io/en/latest/catalog/#colormaps-by-category
                    for available options.
        objects :   list, optional
                    Objects to colorize. If None, will colorize all objects.
        randomize : bool
                    If True (default), will randomly shuffle the colors.

        """
        if objects is None:
            objects = self.objects  # grab once to speed things up

        if not isinstance(palette, cmap._colormap.Colormap):
            palette = cmap.Colormap(palette)

        if randomize:
            # Note: can't use numpy here because it claims array is not 1d
            colors = random.choices(list(palette.iter_colors()), k=len(objects))
        else:
            colors = list(palette.iter_colors(len(objects)))

        colormap = {s: tuple(colors[i].rgba) for i, s in enumerate(objects)}

        self.set_colors(colormap)

    def set_bgcolor(self, c):
        """Set background color.

        Parameters
        ----------
        c :     tuple | str
                RGB(A) color to use for the background.

        """
        self._background.material.set_colors(gfx.Color(c).rgba)

    def _toggle_fps(self):
        """Switch FPS measurement on and off."""
        self.show_fps = not self.show_fps

    def screenshot(
        self, filename="screenshot.png", size=None, pixel_ratio=None, alpha=True
    ):
        """Save a screenshot of the canvas.

        Parameters
        ----------
        filename :      str | pathlib.Path, optional
                        Filename to save to. If ``None``, will return image array.
                        Note that this will always save a PNG file, no matter
                        the extension.
        size :          tuple, optional
                        Size of the screenshot. If provided, will temporarily
                        change the canvas size.
        pixel_ratio :   int, optional
                        Factor by which to scale canvas. Determines image
                        dimensions. Note that this seems to have no effect
                        on offscreen canvases.
        alpha :         bool, optional
                        If True, will export transparent background.

        """
        im = self._screenshot(alpha=alpha, size=size, pixel_ratio=pixel_ratio)
        if filename:
            filename = Path(filename)
            if filename.suffix != ".png":
                filename = filename.parent / f"{filename.name}.png"
            png.from_array(
                im.reshape(im.shape[0], im.shape[1] * im.shape[2]), mode="RGBA"
            ).save(str(filename.resolve()))
        else:
            return im

    def _screenshot(self, alpha=True, size=None, pixel_ratio=None):
        """Return image array for screenshot."""
        if alpha:
            vis = self._background.visible
            self._background.visible = False
        if size:
            os = self.size
            self.size = size
        if pixel_ratio:
            opr = self.renderer.pixel_ratio
            self.renderer.pixel_ratio = pixel_ratio

        # If this is an offscreen canvas, we need to manually trigger a draw first
        # Note: this has to happen _after_ adjust parameters!
        if isinstance(self.canvas, OffscreenRenderCanvas):
            self.canvas.draw()
        else:
            # This is a bit of a hack to make sure a new frame with the (potentially)
            # updated size, pixel ratio, etc. is drawn before taking the screenshot.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self.canvas._subwidget._draw_frame_and_present()

        try:
            im = self.renderer.snapshot()
        except BaseException:
            raise
        finally:
            if alpha:
                self._background.visible = vis
            if size:
                self.size = os
            if pixel_ratio:
                self.renderer.pixel_ratio = opr

        return im

    def set_view(self, view):
        """(Re-)set camera position.

        Parameters
        ----------
        view :      XY | XZ | YZ | dict
                    View to set. Can be inverted to e.g. "-XY" to show view from back.
                    If a dictionary, should describe the state of the camera. Typically,
                    this is obtained by calling `viewer.get_view()`.

        """
        if view == "XY":
            self.camera.show_object(
                self.scene, view_dir=(0.0, 0.0, 1.0), up=(0.0, -1.0, 0.0)
            )
        elif view == "-XY":
            self.camera.show_object(
                self.scene, view_dir=(0.0, 0.0, -1.0), up=(0.0, -1.0, 0.0)
            )
        elif view == "XZ":
            self.camera.show_object(
                self.scene, scale=1, view_dir=(0.0, 1.0, 0.0), up=(0.0, 0.0, 1.0)
            )
        elif view == "-XZ":
            self.camera.show_object(
                self.scene, scale=1, view_dir=(0.0, -1.0, 0.0), up=(0.0, 0.0, 1.0)
            )
        elif view == "YZ":
            self.camera.show_object(
                self.scene, scale=1, view_dir=(1.0, 0.0, 0.0), up=(0.0, -1.0, 0.0)
            )
        elif view == "YZ":
            self.camera.show_object(
                self.scene, scale=1, view_dir=(-1.0, 0.0, 0.0), up=(0.0, -1.0, 0.0)
            )
        elif isinstance(view, dict):
            self.camera.set_state(view)
        else:
            raise TypeError(f"Unable to set view from {type(view)}")

    def get_view(self):
        """Get current camera position."""
        return self.camera.get_state()

    def bind_key(self, key, func, modifiers=None):
        """Bind a function to a key press.

        Note that any existing keybindings for `key` + `modifiers` will be
        silently overwritten.

        Parameters
        ----------
        key :       str
                    Key to bind to. Can be any key on the keyboard.
        func :      callable
                    Function to call when key is pressed.
        modifiers : str | list thereof, optional
                    Modifier(s) to use with the key. Can be "Shift", "Control",
                    "Alt" or "Meta".

        """
        if not callable(func):
            raise TypeError("`func` needs to be callable")

        if not isinstance(key, str):
            raise TypeError(f"Expected `key` to be a string, got {type(key)}")

        if modifiers is None:
            self._key_events[key] = func
        else:
            # We need to make `modifiers` is hashable
            if isinstance(modifiers, str):
                modifiers = (modifiers,)
            elif isinstance(modifiers, (set, list)):
                modifiers = tuple(modifiers)

            if not isinstance(modifiers, tuple):
                raise TypeError(
                    f"Unexpected datatype for `modifiers`: {type(modifiers)}"
                )

            self._key_events[(key, modifiers)] = func


def handle_object_event(event, viewer, actions):
    """Handle object events."""
    # Parse the object (this will be e.g. a Mesh visual)
    obj = event.pick_info["world_object"]

    # Get the ID of the object
    new_hover_id = [k for k, v in viewer.objects.items() if obj in v]
    new_hover_id = new_hover_id[0] if new_hover_id else None

    if new_hover_id:
        if "hide" in actions:
            viewer.hide_objects(new_hover_id)
        if "unhide" in actions:
            viewer.unhide_objects(new_hover_id)
        if "highlight" in actions:
            viewer.highlight_objects(new_hover_id)
        if "unhighlight" in actions:
            viewer.unhighlight_objects(new_hover_id)
        if "pin" in actions:
            viewer.pin_objects(new_hover_id)
        if "unpin" in actions:
            viewer.unpin_objects(new_hover_id)
        if "remove" in actions:
            viewer.remove_objects(new_hover_id)
        if "select" in actions:
            if new_hover_id in viewer.selected:
                viewer.selected = [i for i in viewer.selected if i != new_hover_id]
            else:
                viewer.selected = np.append(viewer.selected, new_hover_id)

        logger.debug(f"Object: {new_hover_id}, Action: {actions}")


def start_ipython_event_loop(gui):
    ip = get_ipython()  # noqa
    if not ip.active_eventloop:
        try:
            ip.enable_gui(gui)
            logger.debug(
                "Looks like you're running in an IPython environment but haven't "
                "started a GUI event loop. We've started one for you using the "
                f"{gui} backend."
            )
        except (ModuleNotFoundError, ImportError):
            logger.warning(
                "Looks like you're running an IPython environment but haven't "
                "started a GUI event loop. We tried to start one for you using the "
                f"{gui} backend (via %gui {gui}) but that failed. If you want a"
                "non-blocking Octarine viewer, you may have to start the event loop "
                "manually (see https://ipython.readthedocs.io/en/stable/config/eventloops.html)."
                "Otherwise just call `Viewer.show()` to start a blocking viewer."
            )
            return False
    return True



import warnings
import numpy as np

from pathlib import Path
from tqdm.auto import trange
from scipy.spatial.transform import Rotation


def make_rotation_video(viewer, video_path, n_frames=100, fps=30, axis="z"):
    """Create a video of the viewer rotating around the specified axis.

    We're using imagio to create the video. Make sure to install if you haven't already:

      `pip install imageio`

    Parameters
    ----------
    viewer :    Viewer
                The viewer to capture frames from.
    video_path : str, optional
                The path to save the video to. Should end with .mp4 or .gif.
                If None, the frames will be returned as a list of numpy arrays
                instead of being saved to a video.
    n_frames :  int
                The number of frames in the video.
    fps :       int
                The frames per second of the video.
    axis :      str
                The axis to rotate around, one of 'x', 'y', or 'z'.

    Returns
    -------
    list of numpy arrays
        If video_path is None, returns a list of frames as numpy arrays. Otherwise, returns `None`.

    """
    if video_path is not None:
        try:
            import imageio
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "imageio is required to create videos. Please install it with `pip install imageio`."
            )

    if isinstance(video_path, str):
        video_path = Path(video_path)

    if isinstance(video_path, Path):
        if video_path.suffix not in (".mp4", ".gif"):
            warnings.warn("video_path should really end with .mp4 or .gif")

    assert axis in ("x", "y", "z"), "Axis must be one of 'x', 'y', or 'z'"
    axis_ix = {"x": 0, "y": 1, "z": 2}[axis]
    frames = []
    orig_orientation = Rotation.from_quat(viewer.get_view().get("rotation", [1, 0, 0, 0]))
    for i in trange(n_frames):  # 180 frames for full rotation
        rot_vec = [0, 0, 0]
        rot_vec[axis_ix] = 2 * np.pi / n_frames * i
        rotation = Rotation.from_rotvec(rot_vec)
        new_orientation = orig_orientation * rotation

        # Set the new rotation
        view = viewer.get_view()
        view["rotation"] = new_orientation.as_quat().tolist()
        viewer.set_view(view)

        # capture screenshot
        frames.append(viewer.screenshot(filename=None, alpha=False))

    if video_path:
        imageio.mimsave(video_path, frames, fps=fps)
    else:
        return frames



import numpy as np
import pygfx as gfx

from functools import wraps

try:
    from PySide6 import QtWidgets, QtCore
except ImportError:
    raise ImportError(
        "Showing controls requires PySide6. Please install it via:\n `pip install PySide6`."
    )

# TODOs:
# - add custom legend formatting (e.g. "{object.name}")
# - show type of object in legend
# - add dropdown to manipulate all selected objects
# - add filter to legend (use item.setHidden(True/False) to show/hide items)
# - highlight object in legend when hovered over in scene
# - make legend tabbed (QTabWidget)


def connect_color_picker(func):
    """Decorator to mark this controls as the active color picker target.

    On some backends (notably macOS) Qt uses a shared native color dialog.
    We therefore route color events through a class-level dispatcher and
    update the active controls instance whenever a color action is triggered.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """Wrapper to activate this controls for subsequent color events."""
        Controls._active_color_controls = self
        return func(self, *args, **kwargs)

    return wrapper


def set_viewer_stale(func):
    """Decorator to set the viewer stale after a function call."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self.viewer._render_stale = True
        return func(self, *args, **kwargs)

    return wrapper


class Controls(QtWidgets.QWidget):
    _shared_color_picker = None
    _active_color_controls = None

    def __init__(self, viewer, width=300, height=400):
        super().__init__()
        self.viewer = viewer
        self.setWindowTitle("Controls")
        self.resize(width, height)

        self.tab_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.tab_layout)

        self.tabs = QtWidgets.QTabWidget()
        self.tabs.setDocumentMode(True)
        self.tabs.setTabPosition(QtWidgets.QTabWidget.West)
        self.tabs.setMovable(True)

        self.tab_layout.addWidget(self.tabs)

        self.tab1 = QtWidgets.QWidget()
        self.tab2 = QtWidgets.QWidget()
        self.tab1_layout = QtWidgets.QVBoxLayout()
        self.tab2_layout = QtWidgets.QVBoxLayout()
        self.tab1.setLayout(self.tab1_layout)
        self.tab2.setLayout(self.tab2_layout)

        self.tabs.addTab(self.tab1, "Legend")
        self.tabs.addTab(self.tab2, "Controls")

        # self.btn_layout = QtWidgets.QVBoxLayout()
        # self.setLayout(self.btn_layout)

        # # Build legend
        # self.build_gui()

        # Build gui
        self.build_legend_gui()
        self.build_controls_gui()

        # Populate legend
        self.update_legend()

        # This determines the target for color changes
        self.active_objects = None
        self.active_volume = None

        self.color_picker = self._get_shared_color_picker()

    @classmethod
    def _get_shared_color_picker(cls):
        """Return the shared QColorDialog instance used by all controls."""
        if cls._shared_color_picker is None:
            picker = QtWidgets.QColorDialog()
            picker.setOption(QtWidgets.QColorDialog.ShowAlphaChannel, on=True)
            picker.currentColorChanged.connect(cls._dispatch_color_changed)
            picker.colorSelected.connect(cls._dispatch_color_selected)
            cls._shared_color_picker = picker
        return cls._shared_color_picker

    @classmethod
    def _dispatch_color_changed(cls, color):
        """Forward color changes to the currently active controls instance."""
        active_controls = cls._active_color_controls
        if active_controls is not None:
            active_controls.set_color(color)

    @classmethod
    def _dispatch_color_selected(cls, *args):
        """Reset active selection for the currently active controls instance."""
        active_controls = cls._active_color_controls
        if active_controls is not None:
            active_controls.reset_active_objects()

    def build_legend_gui(self):
        """Build the legend GUI."""
        # Add legend (i.e. a list widget)
        self.legend = self.create_legend()

        # Add the dropdown to action all selected objects
        self.sel_action = QtWidgets.QPushButton(text="Action")
        self.sel_action_menu = QtWidgets.QMenu(self)
        self.sel_action_menu.addAction("Invert Visibility")
        self.sel_action_menu.actions()[-1].triggered.connect(self.invert_visibility)
        self.sel_action_menu.addAction("Hide All")
        self.sel_action_menu.actions()[-1].triggered.connect(self.hide_all)
        self.sel_action_menu.addAction("Show All")
        self.sel_action_menu.actions()[-1].triggered.connect(self.show_all)
        self.sel_action_menu.addAction("Delete All")
        self.sel_action_menu.actions()[-1].triggered.connect(self.delete_all)
        self.sel_action_menu.addAction("Select All")
        self.sel_action_menu.actions()[-1].triggered.connect(self.select_all)
        self.sel_action_menu.addAction("Clear Selection")
        self.sel_action_menu.actions()[-1].triggered.connect(self.select_none)
        self.sel_action_menu.addAction("Invert Selection")
        self.sel_action_menu.actions()[-1].triggered.connect(self.invert_selection)
        self.sel_action_menu.addAction("Hide selected")
        self.sel_action_menu.actions()[-1].triggered.connect(self.hide_selected)
        self.sel_action_menu.addAction("Show selected")
        self.sel_action_menu.actions()[-1].triggered.connect(self.show_selected)
        self.sel_action_menu.addAction("Delete selected")
        self.sel_action_menu.actions()[-1].triggered.connect(self.delete_selected)
        self.sel_action_menu.addAction("Color selected")
        self.sel_action_menu.actions()[-1].triggered.connect(self.color_selected)
        self.sel_action.setMenu(self.sel_action_menu)
        self.tab1_layout.addWidget(self.sel_action)

    def build_controls_gui(self):
        """Build the legend GUI."""

        # Add dropdown to determine what's happening on hover
        self.on_hover_label = QtWidgets.QLabel("On Hover:")
        self.tab2_layout.addWidget(self.on_hover_label)
        self.on_hover_dropdown = QtWidgets.QComboBox()
        self.on_hover_dropdown.addItems(["Nothing", "Highlight"])
        self.on_hover_dropdown.setToolTip(
            "Action to perform when hovering over an object."
        )
        self.on_hover_dropdown.currentIndexChanged.connect(
            lambda x: setattr(
                self.viewer,
                "on_hover",
                self.on_hover_dropdown.currentText().lower()
                if self.on_hover_dropdown.currentText() != "Nothing"
                else None,
            )
        )
        self.tab2_layout.addWidget(self.on_hover_dropdown)

        # Add dropdown to determine what's happening on double click
        self.on_dclick_label = QtWidgets.QLabel("On Double Click:")
        self.tab2_layout.addWidget(self.on_dclick_label)
        self.on_dclick_dropdown = QtWidgets.QComboBox()
        self.on_dclick_dropdown.addItems(["Nothing", "Hide", "Remove", "Select"])
        self.on_dclick_dropdown.setToolTip(
            "Action to perform when double clicking on an object."
        )
        self.on_dclick_dropdown.currentIndexChanged.connect(
            lambda x: setattr(
                self.viewer,
                "on_double_click",
                self.on_dclick_dropdown.currentText().lower()
                if self.on_dclick_dropdown.currentText() != "Nothing"
                else None,
            )
        )
        self.tab2_layout.addWidget(self.on_dclick_dropdown)

        # Horizontal divider
        self.tab2_layout.addWidget(QHLine())

        # Add button to toggle wireframe
        self.mesh_wireframe_checkbox = QtWidgets.QCheckBox("Wireframe")
        self.mesh_wireframe_checkbox.setChecked(False)

        # Add button to toggle flat shading
        self.mesh_flat_checkbox = self.create_checkbox(
            "Flat Shading", self.tab2_layout, gfx.Mesh, "material.flat_shading"
        )

        # Add button to toggle wireframe
        self.mesh_wireframe_checkbox = QtWidgets.QCheckBox("Wireframe")
        self.mesh_wireframe_checkbox.setChecked(False)

        def set_wireframe(*args):
            for vis in self.viewer.scene.children:
                if isinstance(vis, gfx.Mesh):
                    vis.material.wireframe = self.mesh_wireframe_checkbox.isChecked()

        self.mesh_wireframe_checkbox.toggled.connect(set_wireframe)
        self.tab2_layout.addWidget(self.mesh_wireframe_checkbox)

        # Add button to toggle FPS
        self.fps_checkbox = QtWidgets.QCheckBox("Show FPS")
        self.fps_checkbox.setChecked(self.viewer._show_fps)

        self.fps_checkbox.toggled.connect(
            lambda x: setattr(
                self.viewer,
                "show_fps",
                self.fps_checkbox.isChecked(),
            )
        )
        self.tab2_layout.addWidget(self.fps_checkbox)

        # Horizontal divider
        self.tab2_layout.addWidget(QHLine())

        # Add controls to adjust ambient light
        self.ambient_light_checkbox = QtWidgets.QCheckBox("Ambient Light")
        self.ambient_light_checkbox.setChecked(True)

        def toggle_ambient_light(*args):
            for vis in self.viewer.scene.children:
                if isinstance(vis, gfx.AmbientLight):
                    vis.visible = self.ambient_light_checkbox.isChecked()
                    self.viewer._render_stale = True

        self.ambient_light_checkbox.toggled.connect(toggle_ambient_light)
        self.tab2_layout.addWidget(self.ambient_light_checkbox)

        self.ambient_light_slider = self.create_slider(
            "Intensity",
            0,
            10,
            gfx.AmbientLight,
            "intensity",
            step=0.01,
            parent_layout=self.tab2_layout,
        )

        # Horizontal divider
        self.tab2_layout.addWidget(QHLine())

        # Add dropdown to determine render mode
        self.render_mode_label = QtWidgets.QLabel("Render trigger:")
        self.tab2_layout.addWidget(self.render_mode_label)
        self.render_mode_dropdown = QtWidgets.QComboBox()
        self.render_mode_dropdown.setToolTip(
            "Set trigger for re-rendering the scene. See documentation for details."
        )
        self.render_mode_dropdown.addItems(["Continuous", "Reactive", "Active Window"])
        self.render_mode_dropdown.setItemData(
            0, "Continuously render the scene.", QtCore.Qt.ToolTipRole
        )
        self.render_mode_dropdown.setItemData(
            1,
            "Render only when the scene changes.",
            QtCore.Qt.ToolTipRole,
        )
        self.render_mode_dropdown.setItemData(
            2, "Render only when the window is active.", QtCore.Qt.ToolTipRole
        )
        render_trigger_vals = ["continuous", "reactive", "active_window"]
        self.render_mode_dropdown.currentIndexChanged.connect(
            lambda x: setattr(
                self.viewer,
                "render_trigger",
                render_trigger_vals[self.render_mode_dropdown.currentIndex()],
            )
        )
        # Set default item to whatever the currently set render trigger is
        self.render_mode_dropdown.setCurrentIndex(
            render_trigger_vals.index(self.viewer.render_trigger)
        )
        self.tab2_layout.addWidget(self.render_mode_dropdown)

        # This would make it so the legend does not stretch when
        # we resize the window vertically
        self.tab2_layout.addStretch(1)

    def create_legend(self, spacing=0, index=None):
        """Generate the legend widget."""
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(QtWidgets.QLabel("Legend"))
        list_widget = QtWidgets.QListWidget()
        list_widget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        layout.addWidget(list_widget)
        list_widget.setSpacing(spacing)

        # Add some example items (for debugging only)
        # for i, c in enumerate(["red", "green", "blue"]):
        #     item, item_widget = self.make_legend_entry(f"Item {i}", color=c)
        #     list_widget.addItem(item)
        #     list_widget.setItemWidget(item, item_widget)

        if index is not None:
            self.tab1_layout.insertWidget(index, list_widget)
        else:
            self.tab1_layout.addWidget(list_widget)

        return list_widget

    def make_legend_entry(self, name, color=None, type=None):
        """Generate a legend entry.

        Parameters
        ----------
        name :      str
                    Name of the entry.
        color :     str | tuple | array
                    Color of the entry.

        Returns
        -------
        item :      QtWidgets.QListWidgetItem
                    List item.
        item_widget : QtWidgets.QWidget
                    List item widget.

        """
        # Initialize widget and item
        item_widget = QtWidgets.QWidget()
        item_widget.setObjectName(str(name))
        item = QtWidgets.QListWidgetItem()
        item._id = name  # this helps to identify the item

        # Generate the label
        line_text = QtWidgets.QPushButton(f"{name}", flat=True)
        line_text.setToolTip("Click to select")
        line_text.setProperty("legend_role", "label")
        # Allow long names to shrink first so toggle + color controls stay visible.
        line_text.setSizePolicy(
            QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Preferred
        )
        line_text.setMinimumWidth(0)

        # Generate the checkbox
        line_checkbox = QtWidgets.QCheckBox()
        line_checkbox.setObjectName(str(name))  # this helps to identify the checkbox
        line_checkbox.setMaximumWidth(40)
        line_checkbox.setToolTip("Toggle visibility")
        line_checkbox.setChecked(True)

        def set_property(*args):
            for vis in self.viewer.objects.get(name, []):
                # Navigate to the correct property
                vis.visible = line_checkbox.isChecked()
                self.viewer._render_stale = True
            self.update_legend()

        line_checkbox.toggled.connect(set_property)

        # Generate the button
        if type == gfx.Volume:
            line_push_button = self.create_volume_btn(name, callback=None)
        else:
            line_push_button = self.create_color_btn(name, color=color, callback=None)
        line_push_button.setProperty("legend_role", "control")

        # Generate item layout
        item_layout = QtWidgets.QHBoxLayout()
        item_layout.setContentsMargins(0, 0, 0, 0)  # make layout tight
        item_layout.setSpacing(0)

        # Add text and button to layout
        item_layout.addWidget(line_text)
        item_layout.addWidget(line_checkbox)
        item_layout.addSpacing(15)
        item_layout.addWidget(line_push_button)
        item_layout.setStretch(0, 1)

        # Set layout
        item_widget.setLayout(item_layout)
        item.setSizeHint(item_widget.sizeHint())

        return item, item_widget

    def make_grouped_legend_entry(self, names, group_name=None):
        """Generate a collapsible legend entry for grouped objects.

        Parameters
        ----------
        names : list[str] | tuple[str]
                    Object names that belong to this group.
        group_name : str | None
                    Label shown in the collapsible header.

        Returns
        -------
        item_widget : QtWidgets.QWidget
                    Collapsible widget with one row per grouped member.
        """
        if not names:
            raise ValueError("`names` must contain at least one object name.")

        if group_name is None:
            group_name = f"group ({len(names)})"

        # Outer container that can be used as the list item widget.
        item_widget = QtWidgets.QWidget()
        item_widget.setObjectName(str(group_name))
        item_layout = QtWidgets.QVBoxLayout(item_widget)
        item_layout.setContentsMargins(0, 0, 0, 0)
        item_layout.setSpacing(0)

        # Header button to toggle child visibility.
        header_row = QtWidgets.QWidget()
        header_row_layout = QtWidgets.QHBoxLayout(header_row)
        header_row_layout.setContentsMargins(0, 0, 0, 0)
        header_row_layout.setSpacing(0)

        header = QtWidgets.QToolButton()
        # Prefix with a thin gap so text does not crowd the arrow indicator.
        header.setText(f"  {group_name}")
        header.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        header.setArrowType(QtCore.Qt.RightArrow)
        header.setCheckable(True)
        header.setChecked(False)
        header.setStyleSheet(
            "QToolButton {"
            " text-align: left;"
            " font-size: 13px;"
            " background: transparent;"
            " border: none;"
            " padding: 0px;"
            " }"
            "QToolButton:checked { background: transparent; border: none; }"
            "QToolButton:pressed { background: transparent; border: none; }"
            "QToolButton:hover { background: transparent; border: none; }"
            "QToolButton::menu-indicator { width: 8px; height: 8px; }"
        )
        header.setAutoRaise(True)
        header.setIconSize(QtCore.QSize(8, 8))
        header.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed
        )
        header.setToolTip("Click to expand/collapse grouped legend entries")

        group_checkbox = QtWidgets.QCheckBox()
        group_checkbox.setProperty("legend_role", "group_visibility")
        group_checkbox.setTristate(True)
        group_checkbox.setMaximumWidth(40)
        group_checkbox.setToolTip("Toggle visibility for all group members")

        def toggle_group_visibility(*args):
            is_visible = group_checkbox.isChecked()
            for member_name in names:
                for vis in self.viewer.objects.get(member_name, []):
                    vis.visible = is_visible

                # Mirror the group toggle on each member checkbox without
                # retriggering each member's callback.
                member_checkbox = item_widget.findChild(
                    QtWidgets.QCheckBox, str(member_name)
                )
                if member_checkbox is not None:
                    member_checkbox.blockSignals(True)
                    member_checkbox.setChecked(is_visible)
                    member_checkbox.blockSignals(False)
            self.viewer._render_stale = True

        def coerce_group_checkbox_click_state(*args):
            # Partial is an indicator-only state; user clicks should skip it.
            if group_checkbox.checkState() == QtCore.Qt.PartiallyChecked:
                group_checkbox.blockSignals(True)
                group_checkbox.setCheckState(QtCore.Qt.Checked)
                group_checkbox.blockSignals(False)
                toggle_group_visibility()

        group_checkbox.toggled.connect(toggle_group_visibility)
        group_checkbox.clicked.connect(coerce_group_checkbox_click_state)

        group_color = None
        first_name = names[0]
        if first_name in self.viewer.objects and self.viewer.objects[first_name]:
            try:
                group_color = self.viewer.objects[first_name][0].material.color
            except BaseException:
                group_color = "k"

        group_color_button = self.create_color_btn(
            f"group::{group_name}", color=group_color, callback=None
        )
        group_color_button.setProperty("legend_role", "group_control")
        group_color_button.setToolTip("Click to change color for entire group")
        group_color_button._id = list(names)

        header_row_layout.addWidget(header)
        header_row_layout.addWidget(group_checkbox)
        header_row_layout.addSpacing(15)
        header_row_layout.addWidget(group_color_button)
        header_row_layout.setStretch(0, 1)
        item_layout.addWidget(header_row)

        # Child container with one standard legend row per member.
        content_widget = QtWidgets.QWidget()
        content_layout = QtWidgets.QVBoxLayout(content_widget)
        content_layout.setContentsMargins(18, 0, 0, 0)
        content_layout.setSpacing(0)
        content_widget.setVisible(False)

        for name in names:
            color = None
            vis_type = None
            if name in self.viewer.objects and self.viewer.objects[name]:
                try:
                    color = self.viewer.objects[name][0].material.color
                except BaseException:
                    color = "k"
                vis_type = type(self.viewer.objects[name][0])

            _, child_widget = self.make_legend_entry(name, color=color, type=vis_type)
            child_widget.setProperty("legend_role", "group_member")

            child_layout = child_widget.layout()
            if child_layout is not None:
                child_layout.setContentsMargins(0, 0, 0, 0)
                child_layout.setSpacing(0)

            # Keep rows compact but let Qt compute a non-clipping height.
            child_widget.setSizePolicy(
                QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
            )

            # Apply mild compression: reduce excess row padding without clipping text.
            compact_height = max(20, child_widget.sizeHint().height() - 8)
            child_widget.setMinimumHeight(compact_height)
            child_widget.setMaximumHeight(compact_height)
            content_layout.addWidget(child_widget)

        item_layout.addWidget(content_widget)

        def toggle_group(expanded):
            header.setArrowType(QtCore.Qt.DownArrow if expanded else QtCore.Qt.RightArrow)
            content_widget.setVisible(expanded)

            # Keep parent QListWidgetItem height in sync with expanded/collapsed state.
            if hasattr(self, "legend"):
                for i in range(self.legend.count()):
                    legend_item = self.legend.item(i)
                    if self.legend.itemWidget(legend_item) is item_widget:
                        legend_item.setSizeHint(item_widget.sizeHint())
                        break

        header.toggled.connect(toggle_group)

        return item_widget

    def update_legend(self):
        """Update legend with objects in current scene."""
        # Get visuals in scene
        visuals = self.viewer.visuals

        # Collect ungrouped visuals by object id and grouped visuals by group name.
        object_ids = {}
        grouped_ids = {}
        for vis in visuals:
            if not hasattr(vis, "_object_id"):
                continue

            object_id = vis._object_id
            object_group = getattr(vis, "_object_group", None)

            if object_group is None:
                object_ids[object_id] = object_ids.get(object_id, []) + [vis]
                continue

            grouped_ids[object_group] = grouped_ids.get(object_group, {})
            grouped_ids[object_group][object_id] = (
                grouped_ids[object_group].get(object_id, []) + [vis]
            )

        # Build a unified legend spec for add/update/remove bookkeeping.
        legend_entries = {}
        for object_id, obj_visuals in object_ids.items():
            legend_entries[object_id] = dict(
                kind="single",
                members=[object_id],
                visuals=obj_visuals,
            )
        for group_name, member_dict in grouped_ids.items():
            group_key = f"group::{group_name}"
            members = list(member_dict.keys())
            flattened = []
            for member in members:
                flattened.extend(member_dict[member])
            legend_entries[group_key] = dict(
                kind="group",
                members=members,
                member_visuals=member_dict,
                visuals=flattened,
                group_name=group_name,
            )

        # Go over existing items
        N_items = self.legend.count()
        present = []
        for i in list(range(N_items))[::-1]:
            # Get this item
            item = self.legend.item(i)

            # Clear item if not present anymore
            if item._id not in legend_entries:
                self.legend.takeItem(i)
                continue
            else:
                present.append(item._id)

            entry = legend_entries[item._id]
            item_widget = self.legend.itemWidget(item)

            if entry["kind"] == "single":
                try:
                    color = entry["visuals"][0].material.color
                except BaseException:
                    color = gfx.Color("k")

                line_push_button = next(
                    (
                        button
                        for button in item_widget.findChildren(QtWidgets.QPushButton)
                        if button.property("legend_role") == "control"
                    ),
                    None,
                )
                if line_push_button:
                    line_push_button.setStyleSheet(f"background-color: {color.css}")
            else:
                expected_members = {str(member) for member in entry["members"]}
                current_members = {
                    w.objectName()
                    for w in item_widget.findChildren(QtWidgets.QWidget)
                    if w.property("legend_role") == "group_member"
                }

                # Keep grouped rows in sync when members are added/removed.
                if current_members != expected_members:
                    header = item_widget.findChild(QtWidgets.QToolButton)
                    is_expanded = bool(header and header.isChecked())
                    item_widget = self.make_grouped_legend_entry(
                        entry["members"], group_name=entry["group_name"]
                    )
                    self.legend.setItemWidget(item, item_widget)
                    item.setSizeHint(item_widget.sizeHint())

                    if is_expanded:
                        new_header = item_widget.findChild(QtWidgets.QToolButton)
                        if new_header:
                            new_header.setChecked(True)

                for member_name in entry["members"]:
                    try:
                        color = entry["member_visuals"][member_name][0].material.color
                    except BaseException:
                        color = gfx.Color("k")

                    member_widget = item_widget.findChild(
                        QtWidgets.QWidget, str(member_name)
                    )
                    if not member_widget:
                        continue
                    line_push_button = next(
                        (
                            button
                            for button in member_widget.findChildren(QtWidgets.QPushButton)
                            if button.property("legend_role") == "control"
                        ),
                        None,
                    )
                    if line_push_button:
                        line_push_button.setStyleSheet(
                            f"background-color: {color.css}"
                        )

                group_color_btn = next(
                    (
                        button
                        for button in item_widget.findChildren(QtWidgets.QPushButton)
                        if button.property("legend_role") == "group_control"
                    ),
                    None,
                )
                if group_color_btn:
                    try:
                        group_color = entry["visuals"][0].material.color
                    except BaseException:
                        group_color = gfx.Color("k")
                    group_color_btn.setStyleSheet(
                        f"background-color: {group_color.css}"
                    )

        # Add new items
        for entry_id, entry in legend_entries.items():
            if entry_id not in present:
                try:
                    color = entry["visuals"][0].material.color
                except BaseException:
                    # Note to self: need to make sure we also cater for color arrays
                    # which are in the geometry object
                    color = "k"

                if entry["kind"] == "group":
                    item = QtWidgets.QListWidgetItem()
                    item._id = entry_id
                    item_widget = self.make_grouped_legend_entry(
                        entry["members"], group_name=entry["group_name"]
                    )
                    item.setSizeHint(item_widget.sizeHint())
                else:
                    item, item_widget = self.make_legend_entry(
                        entry_id,
                        color=color,
                        type=type(entry["visuals"][0]),
                    )

                self.legend.addItem(item)
                self.legend.setItemWidget(item, item_widget)

        # Check visibility and selected status of objects
        visible = self.viewer.visible
        for i in range(self.legend.count()):
            item = self.legend.item(i)
            item_widget = self.legend.itemWidget(item)
            entry = legend_entries.get(item._id)

            if entry is None:
                continue

            if entry["kind"] == "group":
                member_ids = entry["members"]

                group_checkbox = next(
                    (
                        checkbox
                        for checkbox in item_widget.findChildren(QtWidgets.QCheckBox)
                        if checkbox.property("legend_role") == "group_visibility"
                    ),
                    None,
                )
                if group_checkbox is not None:
                    visible_count = sum(member_id in visible for member_id in member_ids)
                    is_mixed = 0 < visible_count < len(member_ids)

                    group_checkbox.blockSignals(True)
                    if is_mixed:
                        group_checkbox.setCheckState(QtCore.Qt.PartiallyChecked)
                    elif visible_count > 0:
                        group_checkbox.setCheckState(QtCore.Qt.Checked)
                    else:
                        group_checkbox.setCheckState(QtCore.Qt.Unchecked)

                    if is_mixed:
                        group_checkbox.setToolTip(
                            "Some objects in this group are hidden"
                        )
                    else:
                        group_checkbox.setToolTip(
                            "Toggle visibility for all group members"
                        )
                    group_checkbox.blockSignals(False)
            else:
                member_ids = [item._id]

            for member_id in member_ids:
                line_checkbox = item_widget.findChild(QtWidgets.QCheckBox, str(member_id))
                if line_checkbox:
                    line_checkbox.setChecked(member_id in visible)

                member_widget = item_widget.findChild(QtWidgets.QWidget, str(member_id))
                if member_widget is None:
                    member_widget = item_widget

                line_text = next(
                    (
                        button
                        for button in member_widget.findChildren(QtWidgets.QPushButton)
                        if button.property("legend_role") == "label"
                    ),
                    None,
                )
                if not line_text:
                    continue

                if self.viewer.selected and member_id in self.viewer.selected:
                    line_text.setStyleSheet("color: yellow; text-align: left;")
                else:
                    line_text.setStyleSheet("color: white; text-align: left;")

    @connect_color_picker
    def color_button_clicked(self):
        """Set the active object to be the buttons target."""
        sender = self.sender()
        push_button = self.findChild(QtWidgets.QPushButton, sender.objectName())
        # print(f'click: {push_button.objectName()}')
        self.active_objects = push_button._id
        self.color_picker.show()

    def volume_button_clicked(self):
        """Set the active object to be the buttons target."""
        sender = self.sender()
        push_button = self.findChild(QtWidgets.QPushButton, sender.objectName())
        # print(f'click: {push_button.objectName()}')
        self.active_volume = push_button.objectName()
        self.volume_controls.show()

    @set_viewer_stale
    def set_color(self, color):
        """Color current active object(s). This is the callback for the color picker."""
        if self.active_objects is None:
            return
        elif self.active_objects == "selected":
            targets = self.get_selected()
        elif not isinstance(self.active_objects, (list, tuple)):
            targets = [self.active_objects]
        else:
            targets = list(self.active_objects)

        # Convert QColor to [0-1] RGB
        color = np.array(color.toTuple()) / 255

        self.viewer.set_colors({name: color for name in targets})

    def select_all(self):
        """Select all objects."""
        self.legend.selectAll()

    def select_none(self):
        """Select no objects."""
        self.legend.clearSelection()

    def invert_selection(self):
        """Invert selection."""
        for i in range(self.legend.count()):
            item = self.legend.item(i)
            if item.isSelected():
                self.legend.setItemSelected(item, False)
            else:
                self.legend.setItemSelected(item, True)

    def get_selected(self):
        """Get selected items."""
        sel = []
        for item in self.legend.selectedItems():
            sel.append(item._id)
        return sel

    @connect_color_picker
    def color_selected(self):
        """Set the active object to be the selected objects."""
        self.active_objects = "selected"
        self.color_picker.show()

    @set_viewer_stale
    def hide_all(self):
        """Hide all objects."""
        self.viewer.hide_objects(self.viewer.objects)

    @set_viewer_stale
    def hide_selected(self):
        """Hide selected objects."""
        sel = self.get_selected()
        if sel:
            self.viewer.hide_objects(self.get_selected())

    @set_viewer_stale
    def show_all(self):
        """Show all objects."""
        self.viewer.unhide_objects(None)

    @set_viewer_stale
    def show_selected(self):
        """Show selected objects."""
        sel = self.get_selected()
        if sel:
            self.viewer.unhide_objects(self.get_selected())

    @set_viewer_stale
    def invert_visibility(self):
        """Invert visibility of all objects."""
        vis = self.viewer.visible
        invis = self.viewer.invisible
        self.viewer.hide_objects(vis)
        self.viewer.unhide_objects(invis)

    @set_viewer_stale
    def delete_all(self):
        """Delete all objects."""
        self.viewer.remove_objects(self.viewer.objects)

    @set_viewer_stale
    def delete_selected(self):
        """Delete selected objects."""
        sel = self.get_selected()
        if sel:
            self.viewer.remove_objects(self.get_selected())

    def reset_active_objects(self):
        """Reset active objects."""
        self.active_objects = None

    def create_color_btn(self, name, color=None, callback=None):
        """Generate a colorize button ."""
        # Generate button
        color_btn = QtWidgets.QPushButton()

        # Make sure it doesn't take up too much space
        color_btn.setMaximumWidth(20)
        color_btn.setMaximumHeight(20)
        color_btn.setObjectName(str(name))
        color_btn._id = name  # this helps to identify the associated object

        # Set tooltip
        color_btn.setToolTip("Click to change color")

        # Set color (will be updated subsequently via controls.update_legend())
        if color is None:
            color = "w"
        color = gfx.Color(color)
        color_btn.setStyleSheet(f"background-color: {color.css}")

        # Connect callback (this just sets the active object)
        color_btn.clicked.connect(self.color_button_clicked)

        return color_btn

    def create_volume_btn(self, name, callback=None, color="w"):
        """Generate a button to bring up the volume control."""
        # Generate button
        volume_btn = QtWidgets.QPushButton()

        # Make sure it doesn't take up too much space
        volume_btn.setMaximumWidth(20)
        volume_btn.setMaximumHeight(20)
        volume_btn.setObjectName(str(name))

        # Set tooltip
        volume_btn.setToolTip("Click to adjust volume")

        # Set color (will be updated subsequently via controls.update_legend())
        if color is None:
            color = "w"
        color = gfx.Color(color)
        volume_btn.setStyleSheet(f"background-color: {color.css}")

        # Connect callback (this just sets the active object)
        volume_btn.clicked.connect(self.volume_button_clicked)

        return volume_btn

    def create_checkbox(
        self,
        name,
        parent_layout,
        targets=None,
        property=None,
        callback=None,
        toggle=[],
        index=None,
        default_value=False,
    ):
        """Create a checkbox to toggle a property."""
        checkbox = QtWidgets.QCheckBox(name)

        checkbox.setChecked(bool(default_value))

        def set_property(*args):
            path = property.split(".")
            for vis in self.viewer.scene.children:
                if isinstance(vis, targets):
                    # Navigate to the correct property
                    for p in path[:-1]:
                        vis = getattr(vis, p)
                    setattr(vis, path[-1], checkbox.isChecked())
            for e in toggle:
                e.setEnabled(checkbox.isChecked())
            if callback:
                callback(checkbox.isChecked())

        checkbox.toggled.connect(set_property)

        # set_property()
        if index is not None:
            parent_layout.insertWidget(index, checkbox)
        else:
            parent_layout.addWidget(checkbox)
        return checkbox

    def create_slider(
        self, name, min, max, targets, property, parent_layout, step=1, callback=None
    ):
        """Generate a slider to adjust a property."""
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(QtWidgets.QLabel(name))
        slide = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        slide.setMinimum(min / step)
        slide.setMaximum(max / step)
        # slide.setSingleStep(step)
        val = 0
        for vis in self.viewer.scene.children:
            if isinstance(vis, targets):
                val = getattr(vis, property)
                break
        slide.setValue(val / step)

        if isinstance(step, float):
            val_label = QtWidgets.QLabel(f"{float(val):3.2f}")
        else:
            val_label = QtWidgets.QLabel(f"{int(val):03d}")

        layout.addWidget(val_label)

        def set_value(value):
            value = value * step
            if isinstance(step, float):
                val_label.setText(f"{float(value):3.2f}")
            else:
                val_label.setText(f"{int(value):03d}")

            for target in self.viewer.scene.children:
                if not isinstance(target, targets):
                    continue
                setattr(target, property, value)
            if callback:
                callback(value)

        slide.valueChanged.connect(set_value)

        layout.addWidget(slide)

        parent_layout.addLayout(layout)
        return slide

    def close(self):
        """Close the controls."""
        # Keep shared picker alive but hide it when this controls closes.
        if Controls._active_color_controls is self:
            Controls._active_color_controls = None
        self.color_picker.hide()
        super().close()


class QHLine(QtWidgets.QFrame):
    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)


class QVLine(QtWidgets.QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QtWidgets.QFrame.VLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)



import time
import pytest
import octarine as oc

import trimesh as tm
import numpy as np

# Set random state
np.random.seed(0)


@pytest.fixture
def mesh():
    return tm.creation.icosphere()


@pytest.fixture()
def line_single():
    return np.random.rand(10, 3)


@pytest.fixture()
def line_stack():
    return [np.random.rand(i, 3) for i in np.random.randint(2, 10, 10)]


@pytest.fixture()
def points():
    return np.random.rand(10, 3)


@pytest.fixture()
def points_colors():
    return np.random.rand(10, 3), np.random.rand(10, 3)


def test_adding_generic_objects(mesh, line_single, line_stack, points, points_colors):
    v = oc.Viewer(offscreen=True)

    # Test adding objects generically
    for ob in [mesh, line_single, line_stack, points, points_colors]:
        v.add(ob)
        v.clear()

    v.close()


@pytest.mark.parametrize("color", [None, "red", np.random.rand(3)])
def test_adding_mesh(mesh, color):
    v = oc.Viewer(offscreen=True)
    v.add_mesh(mesh, color=color)
    v.close()


def test_showing_messsage():
    v = oc.Viewer(offscreen=True)
    v.show_message("test", color="red")
    v.show_message(None, color="red")
    v.show_message("test", color="red", duration=1)
    time.sleep(2)
    v.close()



"""
A selection gizmo for the 3d viewer.
"""

import numpy as np
import pylinalg as la
import pygfx as gfx

from pygfx.objects import WorldObject
from pygfx.utils.viewport import Viewport
from pygfx.utils.transform import AffineTransform

from .visuals import lines2gfx


class BaseSelectionGizmo(WorldObject):
    """Gizmo to manage a Selection Box.

    To invoke the Gizmo shift-click on the canvas (see also `modifier`) and start dragging
    to draw a selection rectangle.

    Parameters
    ----------
    viewer :  Viewer
        The viewer to which the gizmo is attached.
    modifier : str
        The modifier key to use to activate the gizmo. Default "Shift".
    outline_color : str, optional
        The color to use for the edge of the selection box. Set to None to disable outline.
    outline_color : str, optional
        The color to use for the fill of the selection box. Set to None to disable fill.
    outline_width : int
        The width of the selection box lines.
    outline_style : str
        The style of the selection box lines.
    force_square : bool
        Whether to force the selection box to be square.
    show_info : bool
        Whether to show a small box with additional infos on the selection box.
    debug : bool
        Whether to print debug information.
    leave : bool
        Whether to leave the selection box after the selection is done.
    callback_after : callable
        A function to call after the selection is done.
    callback_during : callable
        A function to call during the selection.
    name : str
        An identifier for the gizmo.

    """

    def __init__(
        self,
        viewer,
        modifier="Shift",
        outline_color="w",
        outline_opacity=0.7,
        outline_width=1,
        outline_style="dashed",
        fill_color=None,
        fill_opacity=0.3,
        force_square=False,
        show_info=False,
        info_font_size=0.1,
        debug=False,
        leave=False,
        callback_after=None,
        callback_during=None,
        name="SelectionGizmo",
    ):
        assert modifier in ("Shift", "Ctrl", "Alt", None)

        super().__init__()

        self._viewer = viewer
        self._viewport = Viewport.from_viewport_or_renderer(self._viewer.renderer)
        self._camera = self._viewer.overlay_camera  # this is the NDC camera
        self._scene = self._viewer.overlay_scene  # we draw on the overlay scene
        self._scene.add(self)
        self._ndc_to_screen = None
        self._add_default_event_handlers()  # this sets up the event handlers

        # Init - N.B. we're prefixing _ to avoid conflicts with WorldObject properties
        self._debug = debug
        self._show_info = show_info
        self._info_font_size = info_font_size
        self._outline_style = outline_style
        self._outline_width = outline_width
        self._modifier = modifier
        self._outline_color = outline_color
        self._outline_opacity = outline_opacity
        self._fill_color = fill_color
        self._fill_opacity = fill_opacity
        self._force_square = force_square
        self._leave = leave
        self._callback_after = callback_after
        self._callback_during = callback_during
        self._name = name

        self.visible = False  # N.B. this is the WorldObject visibility (no underscore)!
        self._active = False
        self._disable = False  # Can be set to True to (temporarily) disable the gizmo
        self._sel_info = {}
        self._elements = {}

        # Generate the visuals
        self.create_visuals()

    @property
    def bounds(self):
        """Return bounds based on selection.

        Importantly, this always returns the sorted bounds while
        self._sel_info may not be sorted.
        """
        if not self._sel_info:
            return None
        sel = np.vstack([self._sel_info["start"], self._sel_info["end"]])
        return np.vstack([sel.min(axis=0), sel.max(axis=0)])

    def __del__(self):
        """Clean-up."""
        self._clear()
        super().__del__()

    def _clear(self):
        """Remove self from viewer."""
        try:
            self._remove_default_event_handlers()
        except KeyError:
            pass
        self._scene.remove(self)

    def create_visuals(self):
        """Create selection elements."""
        self._visual_fill = None
        self._visual_outline = None

        # Generate fill
        if self._fill_color:
            self._visual_fill = gfx.Mesh(
                gfx.Geometry(
                    positions=np.array(
                        [(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0), (0, 0, 0)],
                        dtype=np.float32,
                    ),
                    indices=np.array([[0, 1, 2, 3]], dtype=np.int32),
                ),
                gfx.MeshBasicMaterial(color=self._fill_color),
            )
            self._visual_fill.material.opacity = self._fill_opacity
            self.add(self._visual_fill)

        # Generate outline
        if self._outline_color:
            self._visual_outline = lines2gfx(
                np.array([(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0), (0, 0, 0)]),
                self._outline_color,
                dash_pattern=self._outline_style,
                linewidth=self._outline_width,
            )
            self._visual_outline.material.opacity = self._outline_opacity
            self.add(self._visual_outline)

        # Generate info text
        if self._show_info:
            self._show_info = (
                gfx.Text(
                    markdown="",
                    font_size=self._info_font_size,
                    anchor="bottomright",
                    material=gfx.TextMaterial(
                        color=self._outline_color
                        if self._outline_color
                        else self._fill_color
                    ),
                ),
                gfx.Text(
                    markdown="",
                    font_size=self._info_font_size,
                    anchor="topleft",
                    material=gfx.TextMaterial(
                        color=self._outline_color
                        if self._outline_color
                        else self._fill_color
                    ),
                ),
            )
            self.add(self._show_info[0])
            self.add(self._show_info[1])

    def _update_gizmo(self, event):
        """Update the transform."""
        if event.type != "before_render":
            return

        # Only update if gizmo is actually active
        if self._active and not self._disable:
            self._update_ndc_screen_transform()

    def _update_ndc_screen_transform(self):
        """Update the NDC (Normalized Device Coordinates) to screen transform."""
        # Note: screen origin is at top left corner of NDC with Y-axis pointing down
        x_dim, y_dim = self._viewport.logical_size
        screen_space = AffineTransform()
        screen_space.position = (-1, 1, 0)
        screen_space.scale = (2 / x_dim, -2 / y_dim, 1)
        self._ndc_to_screen = screen_space.inverse_matrix
        self._screen_to_ndc = screen_space.matrix

    def _screen_to_world(self, pos):
        """Translate screen position to world coordinates."""
        if not self._viewport.is_inside(*pos):
            return None

        # Get position relative to viewport
        pos_rel = (
            pos[0] - self._viewport.rect[0],
            pos[1] - self._viewport.rect[1],
        )

        vs = self._viewport.logical_size

        # Convert position to NDC
        x = pos_rel[0] / vs[0] * 2 - 1
        y = -(pos_rel[1] / vs[1] * 2 - 1)
        pos_ndc = (x, y, 0)

        pos_ndc += la.vec_transform(
            self._camera.world.position, self._camera.camera_matrix
        )
        pos_world = la.vec_unproject(pos_ndc[:2], self._camera.camera_matrix)

        return pos_world

    def _add_default_event_handlers(self):
        """Register Gizmo callbacks."""
        # Add handler for gizmo events
        self._viewport.renderer.add_event_handler(
            self._process_event,
            "pointer_down",
            "pointer_move",
            "pointer_up",
        )
        # Add handler to update gizmo (i.e. the transform)
        self._viewport.renderer.add_event_handler(self._update_gizmo, "before_render")

    def _remove_default_event_handlers(self):
        """Remove Gizmo callbacks."""
        self._viewport.renderer.remove_event_handler(
            self._process_event,
            "pointer_down",
            "pointer_move",
            "pointer_up",
        )
        self._viewport.renderer.remove_event_handler(
            self._update_gizmo, "before_render"
        )

    def _process_event(self, event):
        """Callback to handle gizmo-related events."""
        # If gizmo is disabled, just return
        if self._disable:
            return

        # Triage over event type
        has_mod = self._modifier is None or (self._modifier in event.modifiers)
        if event.type == "pointer_down" and has_mod:
            self._selection_start(event)
            self._viewport.renderer.request_draw()
        elif event.type == "pointer_up" and self._active:
            self._selection_stop(event)
            self._viewport.renderer.request_draw()
        elif event.type == "pointer_move" and self._active:
            self._selection_move(event)
            self._viewport.renderer.request_draw()

    def _selection_start(self, event):
        """Initialize the drag."""
        # Set the positions of the selection rectangle
        world_pos = self._screen_to_world((event.x, event.y))

        if world_pos is None:
            return

        # Set the rectangle to visible
        self.visible = True
        self._active = True
        self._event_modifiers = event.modifiers  # store the modifiers

        if self._visual_outline:
            self._visual_outline.geometry.positions.data[:, 0] = world_pos[0]
            self._visual_outline.geometry.positions.data[:, 1] = world_pos[1]
            self._visual_outline.geometry.positions.update_range()
        if self._visual_fill:
            self._visual_fill.geometry.positions.data[:, 0] = world_pos[0]
            self._visual_fill.geometry.positions.data[:, 1] = world_pos[1]
            self._visual_fill.geometry.positions.update_range()

        # In debug mode we will add points
        if self._debug:
            print("Starting at ", world_pos)
            self.remove(*[c for c in self.children if isinstance(c, gfx.Points)])
            point = gfx.Points(
                gfx.Geometry(
                    positions=np.array(
                        [[world_pos[0], world_pos[1], 0]], dtype=np.float32
                    )
                ),
                material=gfx.PointsMaterial(color="r", size=10),
            )
            self.add(point)

        # Store the selection box coordinates
        self._sel_info = {
            "start": world_pos,
            "end": world_pos,
        }

        # Update info text (if applicable)
        self._update_info()

    def _selection_stop(self, event):
        """Stop the drag on pointer up."""
        # Set the rectangle to invisible
        self._active = False
        if not self._leave:
            self.visible = False

        if self._debug:
            world_pos = self._screen_to_world((event.x, event.y))
            point = gfx.Points(
                gfx.Geometry(
                    positions=np.array(
                        [[world_pos[0], world_pos[1], 0]], dtype=np.float32
                    )
                ),
                material=gfx.PointsMaterial(color="g", size=10),
            )
            self.add(point)
            print("Stopping with Selection box: ", self._sel_info)

        if self._callback_after:
            self._callback_after(self)

    def _selection_move(self, event):
        """Move the selection box with the cursor."""
        # Set the positions of the rectangle
        world_pos = self._screen_to_world((event.x, event.y))

        # If we're outside viewport, return
        if world_pos is None:
            return

        # Make square if necessary
        if self._force_square:
            dx, dy, dz = world_pos - self._sel["start"]
            dmin = min(abs(dx), abs(dy))
            world_pos[0] = self._sel_info["start"][0] + np.sign(dx) * dmin
            world_pos[1] = self._sel_info["start"][1] + np.sign(dy) * dmin

        if self._visual_outline:
            # The first and the last point on the line remain on the origin
            # The second point goes to (origin, new_y), the third to (new_x, new_y)
            # The fourth to (new_x, origin)
            self._visual_outline.geometry.positions.data[1, 1] = world_pos[1]
            self._visual_outline.geometry.positions.data[2, 0] = world_pos[0]
            self._visual_outline.geometry.positions.data[2, 1] = world_pos[1]
            self._visual_outline.geometry.positions.data[3, 0] = world_pos[0]
            self._visual_outline.geometry.positions.update_range()

        if self._visual_fill:
            self._visual_fill.geometry.positions.data[1, 1] = world_pos[1]
            self._visual_fill.geometry.positions.data[2, 0] = world_pos[0]
            self._visual_fill.geometry.positions.data[2, 1] = world_pos[1]
            self._visual_fill.geometry.positions.data[3, 0] = world_pos[0]
            self._visual_fill.geometry.positions.update_range()

        # Store the selection box coordinates
        self._sel_info["end"] = world_pos

        # Update info text (if applicable)
        self._update_info()

        if self._debug:
            print(f"{self._name} moving to {world_pos}")
            point = gfx.Points(
                gfx.Geometry(
                    positions=np.array(
                        [[world_pos[0], world_pos[1], 0]], dtype=np.float32
                    )
                ),
                material=gfx.PointsMaterial(color="w", size=10),
            )
            self.add(point)

    def _update_info(self):
        """Update the info text."""
        if not self._show_info:
            return

        if not self._sel_info:
            return

        self._show_info[0].set_text(
            f"({self._sel_info['start'][0]:.2f}, {self._sel_info['start'][1]:.2f})"
        )
        self._show_info[1].set_text(
            f"({self._sel_info['end'][0]:.2f}, {self._sel_info['end'][1]:.2f})"
        )

        self._show_info[0].local.position = self._sel_info["start"]
        self._show_info[1].local.position = self._sel_info["end"]


# We're wrapping the BaseSelectionGizmo in another class to provide a more user-friendly interface
class SelectionGizmo:
    """Selection gizmo for the 3D viewer.

    This gizmo allows to select objects in the 3d viewer by drawing a rectangle.

    Parameters
    ----------
    viewer : Viewer
        The viewer to which the gizmo is attached.
    callback : callable, optional
        A function to call when the selection is done. Must accept the selection as a dictionary (see
        [`SelectionGizmo._find_selected_objects`][octarine.selection.SelectionGizmo._find_selected_objects]
        for details). Can also be added post-hoc using `SelectionGizmo.add_callback().
    ignore_invisible :  bool
        Whether to ignore invisible objects. Default is True.
    **kwargs
        Keyword arguments passed to BaseSelectionGizmo.

    """

    def __init__(self, viewer, callback=None, ignore_invisible=True, **kwargs):
        self.callbacks = []
        if not isinstance(callback, type(None)):
            self.add_callback(callback)

        if "callback_after" in kwargs:
            raise ValueError(
                "The `callback_after` keyword argument is reserved for internal use!"
            )
        kwargs["callback_after"] = self.handle_post_selection
        self.gizmo = BaseSelectionGizmo(viewer=viewer, **kwargs)
        self.ignore_invisible = ignore_invisible

    def add_callback(self, callback):
        if not callable(callback):
            raise ValueError("`callback` must be a single callable")
        self.callbacks.append(callback)

    def handle_post_selection(self, gizmo):
        """Handle things after selection."""
        # If no callback is set, we're done here
        if not self.callbacks and not self.gizmo._debug:
            return

        # Find the selected objects
        selection = self.find_selected_objects()
        if self.gizmo._debug:
            n_clipped = len([k for k in selection if selection[k]["clipped"]])
            n_contained = len([k for k in selection if selection[k]["contained"]])
            print(f"Selected objects: {n_clipped} clipped, {n_contained} contained:")
            print(selection)

        # Run the callbacks
        for func in self.callbacks:
            func(selection)

    def find_selected_objects(self):
        """Find objects that are selected by the selection box.

        Notes
        -----
        - currently only Meshes and Points are supported; other objects are returns as `None`

        Returns
        -------
        dict
                Returns a dictionary matching `Viewer.objects`, i.e. keyed by object groups:

                    {
                        "group_name": {
                            "clipped": bool,  # whether the object(s) is at least clipped
                            "contained": bool,  # whether the object(s) is fully contained
                            "objects": [  # an entry for each object in the object group
                                {
                                "clipped": bool,
                                "contained": bool,
                                "data": np.ndarray | None (if not clipped or fully contained)
                                }, ...
                            ]
                        }
                    }

        """
        # If no selection box, return right away
        if not self.gizmo._sel_info:
            return {}

        # Next, we need to iterate over all objects, (1) project their data to screen space
        # and (2) determine whether they are inside the selection box
        selection = {}
        for name in (
            self.gizmo._viewer.visible
            if self.ignore_invisible
            else self.gizmo._viewer.objects
        ):
            selection[name] = {"clipped": False, "contained": True, "objects": []}
            objects = self.gizmo._viewer.objects[name]
            for ob in objects:
                # Extract the data depending on the type of object
                if not isinstance(ob, (gfx.Mesh, gfx.Points, gfx.Line)):
                    # Note to self: we could use object boundaries where no data is available
                    if self.gizmo._debug:
                        print(f"Object {ob} not supported")
                    selection[name]["contained"] = None
                    selection[name]["objects"].append(
                        {"clipped": None, "contained": None, "data": None}
                    )
                    continue

                # Get the data representing the object
                data = ob.geometry.positions.data

                # Apply world transform
                data = la.vec_transform(data, ob.world.matrix)

                # Project the data to screen space
                # (N.B. were using the viewer's normal camera, not the overlay's NDC camera)
                data_screen = la.vec_transform(
                    data, self.gizmo._viewer.camera.camera_matrix
                )

                # Check whether the object is in the selection box
                mask = self.is_data_in_selection_box(data_screen)

                # Store the results
                is_clipped = bool(np.any(mask))  # avoid getting np.True_/np.False_

                if isinstance(ob, gfx.Line):
                    # Lines can have breaks where the data will be `nan` - these will
                    # always count as not inside the selection box. We need to ignore
                    # these when checking for full containment.
                    data_not_nan = np.all(~np.isnan(data), axis=1)
                    is_contained = bool(np.all(mask[data_not_nan]))
                else:
                    is_contained = bool(np.all(mask))

                # If the object is either fully contained or fully outside, we don't
                # really need to pass the mask
                if is_contained or not is_clipped:
                    mask = None

                selection[name]["objects"].append(
                    {"clipped": is_clipped, "contained": is_contained, "data": mask}
                )
                # Propagate results to parent group
                if not is_contained:
                    selection[name]["contained"] = False
                if is_clipped:
                    selection[name]["clipped"] = True

        return selection

    def is_data_in_selection_box(self, data_screen):
        """Check whether selected obejct is inside selection box.

        Parameters
        ----------
        data_screen : (N, 3) array
            The data in screen coordinates.

        Returns
        -------
        bool
            Whether the object is inside the selection box.

        """
        if not self.gizmo._sel_info:
            return np.zeros(data_screen.shape[0], dtype=bool)

        # Check whether the object is in the selection box
        return np.all(
            (data_screen[:, [0, 1]] >= self.gizmo.bounds[0, :-1])
            & (data_screen[:, [0, 1]] <= self.gizmo.bounds[1, :-1]),
            axis=1,
        )
